"""Delete all runs and metadata for an experiment"""

import os
import sys
from pathlib import Path
from argparse import ArgumentParser

# --- Constants
CSV = "csvs"
LOG = "logs"
SCR = "scripts"


def verify():
    response = input("\nDelete this experiment? This is NOT REVERSIBLE! [y, n]: ")
    if response.lower() == "y":
        return True
    return False


def exists(path):
    if Path(path).exists():
        return True
    return False


def experiment_deleter(csv_path):
    """Function to delete all run data associated with an experiment
    Args:
        csv_path (str): path to a hyperparam or result csv
    """

    csv_path = Path(csv_path).resolve()
    csv_name = "_".join(csv_path.stem.split("_")[:-1])

    root = csv_path.parent.parent

    csv_path = root / CSV / f"{csv_name}*"
    log_path = root / LOG / csv_name
    scr_path = root / SCR / csv_name

    cmd = "rm -rf "
    cmd += str(csv_path) + " "
    cmd += str(log_path) + " "
    cmd += str(scr_path)

    if exists(csv_path) or exists(log_path) or exists(scr_path):
        print("\n- command: {}\n".format(cmd))

        if verify():
            os.system(cmd)


def parser(args):
    p = ArgumentParser(
        description="- remove all results, logs, checkpoints, and scripts associated with a given experiment"
    )
    p.add_argument(
        "--csv_path",
        type=str,
        required=True,
        help="- path to specific experiment ie. csv_path=exp/xr_pna/csvs/2021-11-03_19-10-43_PDT_hyper.csv",
    )
    parsed = p.parse_args(args)
    arguments = {name: getattr(parsed, name) for name in vars(parsed)}

    return arguments


def delete(csv_path, **kwargs):
    experiment_deleter(csv_path)


if __name__ == "__main__":
    args = parser(sys.argv[1:])
    delete(**args)



"""Interface for getting available training callbacks"""

from tfcaidm.common.inherit import inherit
from tfcaidm.train.custom import _zoo


@inherit(_zoo.calls)
def available_callbacks():
    return custom_callback()


# --- callbacks in csv
def custom_callback():
    customs = {}
    return customs


# --- example trainer hyperparams
def sample_params():
    return {
        "train": {
            "trainer": {
                "seed": 0,
                "n_folds": 1,
                "batch_size": 8,
                "iters": 3000,
                "steps": 100,
                "valid_freq": 6,
                "lr": 0.0003,
                "lr_alpha": 0.25,
                "lr_decay": 0.97,
                "callbacks": None,
            }
        }
    }



"""Defined loss functions"""

import tfcaidm.losses.funcs.entropy as entropy
import tfcaidm.losses.funcs.distance as distance
import tfcaidm.losses.funcs.dice as dice
import tfcaidm.losses.funcs.focal as focal
import tfcaidm.losses.funcs.tversky as tversky

losses = {
    "sce": entropy.SparseCategoricalCrossentropy,
    "wce": entropy.WeightedCategoricalCrossentropy,
    "mae": distance.MeanAbsoluteError,
    "mse": distance.MeanSquaredError,
    "dice": dice.DiceLoss,
    "logcosh_dice": dice.LogCoshDiceLoss,
    "tversky": tversky.TverskyLoss,
    "focal_tversky": tversky.FocalTverskyLoss,
    "focal": focal.FocalLoss,
}



"""Dimensionality transformation layers"""

import tensorflow as tf
from tensorflow.keras import layers


# --- Change feature map shape in the (D, W, H) dim.


def average_pool(x, s=(1, 2, 2), **kwargs):
    x = layers.AveragePooling3D(pool_size=s, strides=s, padding="same")(x)

    return x


def max_pool(x, s=(1, 2, 2), **kwargs):
    x = layers.MaxPooling3D(pool_size=s, strides=s, padding="same")(x)

    return x


def up_sample(x, s=(1, 2, 2)):
    return layers.UpSampling3D(size=s)(x)


def global_max_pool(x, keepdims=False):
    if tf.__version__ >= "2.6":
        return layers.GlobalMaxPooling3D(keepdims=keepdims)(x)
    else:
        pool = layers.GlobalMaxPooling3D()(x)
        if keepdims:
            return layers.Reshape((1, 1, 1, x.shape[-1]))(pool)
        return pool


def global_average_pool(x, keepdims=False):
    if tf.__version__ >= "2.6":
        return layers.GlobalAveragePooling3D(keepdims=keepdims)(x)
    else:
        pool = layers.GlobalAveragePooling3D()(x)
        if keepdims:
            return layers.Reshape((1, 1, 1, x.shape[-1]))(pool)
        return pool


# --- Change feature map shape in the (C) dim.


def broadcast(func, *args, **kwargs):
    if len(*args) > 1:
        return func(**kwargs)(*args)
    else:
        return args[0][0]


add = lambda *args, **kwargs: broadcast(tf.keras.layers.Add, *args, **kwargs)
avg = lambda *args, **kwargs: broadcast(tf.keras.layers.Average, *args, **kwargs)
concat = lambda *args, **kwargs: broadcast(tf.keras.layers.Concatenate, *args, **kwargs)


# --- Reorganize dimensions


def expand(in_feature, out_feature):
    i = len(in_feature.shape)
    o = len(out_feature.shape)
    for _ in range(i, o):
        in_feature = tf.expand_dims(in_feature, axis=-2)
    return in_feature


flatten = lambda x: layers.Flatten()(x)



"""Model visualization tools"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.ndimage import zoom

import tensorflow as tf
from tensorflow.keras import activations, models
from jarvis.utils.display import montage


class Base:
    def __init__(self, model, **kwargs):
        self.model = model

    def get_layer(self, layer):
        if type(layer) == int:
            hidden = self.model.layers[layer]
        elif type(layer) == str:
            hidden = self.model.get_layer(layer)
        else:
            raise ValueError(
                "ERROR! `layer` must be either a `string` or `int` denoting the layer name or layer index respectively."
            )

        return hidden

    def get_grad_model(self, layer, output_name):
        grad_model = models.Model(
            [self.model.inputs],
            {
                "hidden": self.get_layer(layer).output,
                "output": self.model.output[output_name],
            },
        )

        return grad_model

    def norm(self, x):
        return activations.relu(x) / tf.reduce_max(x, keepdims=True)

    def heatmap(self, x, layer, output_name, class_of_interest=1, **kwargs):
        grad_cam = self.get_grad_cam(x, layer, output_name, class_of_interest)
        heat_map = self.norm(grad_cam)

        return heat_map


class GradCAM(Base):
    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

    def get_grad_cam(self, x, layer, output_name, class_of_interest=1, **kwargs):
        grad_model = self.get_grad_model(layer, output_name)

        with tf.GradientTape() as tape:
            model_outputs = grad_model(x)
            hidden = model_outputs["hidden"]
            output = model_outputs["output"]
            cam = output[..., class_of_interest]

        grad = tape.gradient(cam, hidden)
        axis = [*range(len(grad.shape[:-1]))]
        pool_grad = tf.reduce_mean(grad, axis=axis)
        grad_cam = hidden @ pool_grad[..., tf.newaxis]

        return grad_cam


class GradCAMpp(Base):
    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

    def get_grad_cam(self, x, layer, output_name, class_of_interest=1, **kwargs):
        grad_model = self.get_grad_model(layer, output_name)

        with tf.GradientTape() as tape:
            model_outputs = grad_model(x)
            hidden = model_outputs["hidden"]
            output = model_outputs["output"]
            cam = output[..., class_of_interest]

        grad = tape.gradient(cam, hidden)
        grad = activations.relu(grad) / grad
        axis = [*range(len(grad.shape[:-1]))]
        pool_grad = tf.reduce_sum(grad, axis=axis)
        grad_cam = hidden @ pool_grad[..., tf.newaxis]

        return grad_cam


def overlay(x, heatmap, cmap="jet", alpha=0.2, figsize=(7, 7)):

    # --- Use Jarvis montage function to collapse into a N x N grid
    im = np.squeeze(montage(x))
    hm = np.squeeze(montage(heatmap))

    # --- Zoom
    hm = zoom(hm, zoom=np.array(im.shape) / np.array(hm.shape), order=1)

    # --- Draw figure
    plt.clf()
    plt.figure(figsize=figsize)
    plt.axis("off")
    plt.imshow(im, cmap="gray")
    plt.imshow(hm, cmap=cmap, alpha=alpha)



"""Squeeze and excitation block"""

import math

import tfcaidm.models.layers.activation as activation
import tfcaidm.models.layers.learner as learner
import tfcaidm.models.layers.transform as transform


def se(x, c, hyperparams, r=4, **kwargs):
    """Implementation of squeeze-and-exication module https://arxiv.org/abs/1709.01507"""

    c = x.shape[-1]

    # --- Squeeze (pool and combine feature maps)
    gmp = transform.global_max_pool(x, keepdims=True)
    gap = transform.global_average_pool(x, keepdims=True)

    # --- Excite (convolution over squeezed features)
    conv1 = learner.conv_layer(c=int(math.ceil(c / r)), k=1)
    conv2 = learner.conv_layer(c=c, k=1)
    activ = activation.nonlinearity(hyperparams)
    gpl = conv2(activ(conv1(gmp))) + conv2(activ(conv1(gap)))

    # --- Attention
    scale = activation.attn_msk(gpl, hyperparams)

    return x * scale



"""Accuracy related metrics"""

import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras import metrics


reduce = lambda x: tf.math.reduce_sum(x)


class Accuracy(layers.Layer):
    def __init__(self, name="Accuracy", **kwargs):
        super(Accuracy, self).__init__(name=name, **kwargs)
        self.metric = metrics.Accuracy()

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        class_of_interest=1,
        add_metric=True,
        **kwargs,
    ):

        true = tf.cast(y_true[..., 0] == class_of_interest, tf.float32)
        pred = tf.cast(tf.math.argmax(y_pred, axis=-1) == class_of_interest, tf.float32)

        metric = self.metric(y_true=true, y_pred=pred, sample_weight=weights)

        if add_metric:
            self.add_metric(metric, name=self.metric.name)

        return metric


class BalancedAccuracy(layers.Layer):
    def __init__(self, name="BalancedAccuracy", **kwargs):
        super(BalancedAccuracy, self).__init__(name=name, **kwargs)
        self.metric_name = name

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        class_of_interest=1,
        epsilon=1e-6,
        add_metric=True,
        **kwargs,
    ):

        true = tf.cast(y_true[..., 0] == class_of_interest, tf.float32)
        pred = tf.cast(tf.math.argmax(y_pred, axis=-1) == class_of_interest, tf.float32)

        if weights is not None:
            true *= tf.cast(weights[..., 0] != 0, tf.float32)
            pred *= tf.cast(weights[..., 0] != 0, tf.float32)

        tp = reduce(true * pred)
        fp = reduce((1 - true) * pred)
        fn = reduce(true * (1 - pred))
        tn = reduce((1 - true) * (1 - pred))

        sens = tp / (tp + fn)
        spec = tn / (tn + fp)

        A = sens + spec
        B = 2

        metric = A / B

        if add_metric:
            self.add_metric(metric, name=self.metric_name)

        return metric



"""Atrous convolution blocks"""

import numpy as np

from tfcaidm.models.utils import select as model_select
import tfcaidm.models.layers.transform as transform


def aspp(x, c, k, s, hyperparams):
    """Atrous spatial pyramid pooling https://arxiv.org/pdf/1606.00915.pdf"""

    x = model_select.conv_selection(x=x, c=c, k=1, s=1, hyperparams=hyperparams)

    x_list = [x]
    n_branches = get_branches(hyperparams)

    for i in range(n_branches):
        d = get_dilation_size(x, hyperparams, scale=i + 1)
        x_new = model_select.conv_selection(x=x, c=c, k=k, d=d, hyperparams=hyperparams)
        x_list.append(x_new)

    x = transform.add(x_list)
    x = model_select.conv_selection(x=x, c=c, k=k, s=s, hyperparams=hyperparams)

    return x


def acsp(x, c, k, s, hyperparams):
    """Atrous convolution spatial pooling https://arxiv.org/pdf/1706.05587.pdf"""

    x = model_select.conv_selection(x=x, c=c, k=1, s=1, hyperparams=hyperparams)

    n_branches = get_branches(hyperparams)

    for i in range(n_branches):
        d = get_dilation_size(x, hyperparams, scale=i + 1)
        x = model_select.conv_selection(x=x, c=c, k=k, d=d, hyperparams=hyperparams)

    x = model_select.conv_selection(x=x, c=c, k=k, s=s, hyperparams=hyperparams)

    return x


def wasp(x, c, k, s, hyperparams):
    """Waterfall atrous spatial pooling https://arxiv.org/pdf/1706.05587.pdf"""

    x = model_select.conv_selection(x=x, c=c, k=1, s=1, hyperparams=hyperparams)

    x_list = [x]
    n_branches = get_branches(hyperparams)

    for i in range(n_branches):
        d = get_dilation_size(x, hyperparams, scale=i + 1)
        x = model_select.conv_selection(x=x, c=c, k=k, d=d, hyperparams=hyperparams)
        x_list.append(x)

    x = transform.add(x_list)
    x = model_select.conv_selection(x=x, c=c, k=k, s=s, hyperparams=hyperparams)

    return x


def get_branches(hyperparams):
    return hyperparams["model"]["branches"]


def get_dilation_size(x, hyperparams, scale=1):
    is_3d = True if x.shape[1] > 1 else False

    atrous_rate = hyperparams["model"]["atrous_rate"]
    kernel_size = np.array(hyperparams["model"]["kernel_size"])

    new_kernel = kernel_size * atrous_rate * scale

    if not is_3d:
        return (1, *new_kernel[1:])

    return tuple(new_kernel)



"""Dice score related metrics"""

import tensorflow as tf
from tensorflow.keras import layers

from tfcaidm.common.constants import DELIM


reduce = lambda x: tf.math.reduce_sum(x)


class DiceScore(layers.Layer):
    """
    Computes the dice-score (F1) coefficient metric for a given

    Args:
        y_true (np.ndarray): ground-truth label
        y_pred (np.ndarray): predicted logits scores
        c (int) : class to calculate DSC on

    """

    def __init__(self, name="DiceScore", **kwargs):
        super(DiceScore, self).__init__(name=name, **kwargs)
        self.metric_name = name

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        class_of_interest=1,
        epsilon=1e-6,
        add_metric=True,
        **kwargs,
    ):
        true = tf.cast(y_true[..., 0] == class_of_interest, tf.float32)
        pred = tf.cast(tf.math.argmax(y_pred, axis=-1) == class_of_interest, tf.float32)

        if weights is not None:
            true *= tf.cast(weights[..., 0] != 0, tf.float32)
            pred *= tf.cast(weights[..., 0] != 0, tf.float32)

        A = reduce(true * pred) * 2
        B = reduce(true) + reduce(pred) + epsilon

        metric = A / B

        if add_metric:
            self.add_metric(
                metric, name=(self.metric_name + f"{DELIM}{class_of_interest}")
            )

        return metric



"""Compare all experiments for a given project"""

import os
import sys
import pandas as pd
from pathlib import Path
from argparse import ArgumentParser

# --- Constants
CSV = "csvs"


def experiment_comparison(project_path):
    """Function to compare all experiments associated with a project
    Args:
        project_path (str): path to a hyperparam or result csv
    """

    project_path = Path(project_path).resolve()
    csv_path = project_path / CSV
    csvs = os.listdir(csv_path)

    results = [
        (csv_path / f)
        for f in list(
            filter(
                lambda f: f.split("_")[-1] == "results.csv",
                csvs,
            )
        )
    ]

    try:
        return pd.concat([pd.read_csv(result) for result in results])
    except:
        return pd.DataFrame()


def parser(args):
    p = ArgumentParser(
        description="- compare all experiments associated with a project"
    )
    p.add_argument(
        "--project_path",
        type=str,
        required=True,
        help="- path to a specific project csvs ie. project_path=exp/xr_pna/",
    )
    parsed = p.parse_args(args)
    arguments = {name: getattr(parsed, name) for name in vars(parsed)}

    return arguments


def compare(project_path, **kwargs):
    return experiment_comparison(project_path)


if __name__ == "__main__":
    args = parser(sys.argv[1:])
    compare(**args)



"""Validate and store yaml config"""

import pandas as pd
from pathlib import Path
from ruamel.yaml import YAML


class Config:
    def __init__(self, path, *args, **kwargs):
        """Wrapper class to handle and store the pipeline configuration

        Args:
            config (str): path to a yaml file or dict
        """

        path = str(Path(path).resolve())

        try:
            config = Config.load_yaml(path)
            err = Config.error(config)
            if err:
                assert False, "ERROR! Missing parameters in yaml file!"
        except ValueError:
            print("ERROR! could not load yaml file in <Class Config>!")

        try:
            Config.validate_inputs(path, config)
        except ValueError:
            print("ERROR! Missing inputs or outputs in yaml file!")

        self.config = config

    @staticmethod
    def load_yaml(path):
        with open(path, "r") as f:
            yaml = YAML(typ="safe")
            return yaml.load(f)

    @staticmethod
    def error(config):
        issues = "env" not in config
        issues = issues or ("path" not in config["env"])
        issues = issues or ("name" not in config["env"]["path"])
        issues = issues or ("root" not in config["env"]["path"])
        issues = issues or ("client" not in config["env"]["path"])

        issues = "model" not in config
        issues = issues or ("activ" not in config["model"])
        issues = issues or ("atrous_rate" not in config["model"])
        issues = issues or ("attn_msk" not in config["model"])
        issues = issues or ("bneck" not in config["model"])
        issues = issues or ("branches" not in config["model"])
        issues = issues or ("conv_type" not in config["model"])
        issues = issues or ("dblock" not in config["model"])
        issues = issues or ("depth" not in config["model"])
        issues = issues or ("eblock" not in config["model"])
        issues = issues or ("elayer" not in config["model"])
        issues = issues or ("kernel_size" not in config["model"])
        issues = issues or ("model" not in config["model"])
        issues = issues or ("norm" not in config["model"])
        issues = issues or ("order" not in config["model"])
        issues = issues or ("pool_type" not in config["model"])
        issues = issues or ("strides" not in config["model"])
        issues = issues or ("width" not in config["model"])
        issues = issues or ("width_scaling" not in config["model"])
        issues = issues or ("pool_type" not in config["model"])

        issues = "train" not in config
        issues = issues or ("trainer" not in config["train"])
        issues = issues or ("batch_size" not in config["train"]["trainer"])
        issues = issues or ("callbacks" not in config["train"]["trainer"])
        issues = issues or ("iters" not in config["train"]["trainer"])
        issues = issues or ("lr" not in config["train"]["trainer"])
        issues = issues or ("lr_alpha" not in config["train"]["trainer"])
        issues = issues or ("lr_decay" not in config["train"]["trainer"])
        issues = issues or ("n_folds" not in config["train"]["trainer"])
        issues = issues or ("seed" not in config["train"]["trainer"])
        issues = issues or ("steps" not in config["train"]["trainer"])
        issues = issues or ("valid_freq" not in config["train"]["trainer"])

        return issues

    @staticmethod
    def validate_inputs(path, config):
        # --- Load jarvis client yaml
        client = Config.load_yaml(config["env"]["path"]["client"])

        client_specs = client["specs"]
        assert "xs" in client_specs, "ERROR! Client yaml must have a `xs` field!"
        assert "ys" in client_specs, "ERROR! Client yaml must have a `ys` field!"
        client_params = [*client_specs["xs"].keys(), *client_specs["ys"].keys()]

        train_specs = config["train"]
        assert "xs" in train_specs, "ERROR! Train yaml must have a `xs` field!"
        assert "ys" in train_specs, "ERROR! Train yaml must have a `ys` field!"

        train_params = []
        train_params += get_unique_subfields(train_specs["xs"])
        train_params += get_unique_subfields(train_specs["ys"])

        assert set(client_params) == set(
            train_params
        ), f"""ERROR! Inconsistency between yaml configuration files!
        
        Issues:
            client_params = {set(client_params)}
            train_params = {set(train_params)}

            client_yml = {config["env"]["path"]["client"]}
            train_yml = {path}

        Note: client_params and train_params must be identical"""


def get_unique_subfields(field):
    """Extract all unique subfields from a given top-level field from train yml

    Example:

    field = ys:
                pna:
                    mask:
                        name: lung
                        mask_weight = 10
                    loss = sce
                    metrics = acc
                cov:
                    mask:
                        name: thor
                        mask_weight = 100
                    loss = sce
                    metrics = acc

    returns [pna, lung, cov, thor]

    Args:
        field (type): YAML field name, ex: `xs` or `ys`

    Returns:
        list: list of subfields
    """

    params = []
    for k in field:
        params += [k]
        subfield = field[k]

        if type(subfield) == dict:
            for k in subfield:

                if type(subfield[k]) == dict:
                    for n in subfield[k]:

                        if n == "name":
                            if type(subfield[k][n]) == list:
                                params += subfield[k][n]
                            else:
                                params += [subfield[k][n]]

    return params



"""Client hyperparameter interface"""

from jarvis.train.client import Client

from tfcaidm.common.reproducibility import set_determinism
from tfcaidm.data.utils import class_weights
from tfcaidm.data.utils import positional_encoding
from tfcaidm.jobs.utils.config import get_unique_subfields
from tfcaidm.jobs.utils.params import HyperParameters


# --- Wrapper for extracting client yml inputs and outputs
def get_yml_params(func):
    def extract(self, *args, **kwargs):
        tasks_dict = func(self, *args, **kwargs)
        targets = {}

        # --- Associate inputs and outputs with their shape
        for k, v in tasks_dict:
            shape = v["shape"]["saved"]
            targets[k] = shape

        return targets

    return extract


class JClient(HyperParameters, Client):
    def __init__(self, path, hyperparams, configs={}, *args, **kwargs):
        HyperParameters.__init__(self, hyperparams, *args, **kwargs)
        set_determinism(self.hyperparams["train"]["trainer"]["seed"])
        Client.__init__(
            self,
            path,
            configs=configs,
            custom_layers=True,
            *args,
            **kwargs,
        )

    @get_yml_params
    def get_shapes(self, key):
        return self.specs[key].items()

    def get_input_shapes(self):
        specs = self.get_shapes("xs")
        params = get_unique_subfields(self.hyperparams["train"]["xs"])

        xs = {}

        for param in params:
            for k in specs.keys():
                if k == param:
                    xs[param] = specs[k]

        return xs

    def get_output_shapes(self):
        specs = self.get_shapes("xs")
        params = get_unique_subfields(self.hyperparams["train"]["ys"])

        ys = {}

        for param in params:
            for k in specs.keys():
                if k == param:
                    ys[param] = specs[k]

        return ys

    def dataset_size(self, fold=0):
        df = self.db.header
        train = sum(df["valid"] != fold)
        valid = sum(df["valid"] == fold)

        return {"train": f"{train:,}", "valid": f"{valid:,}"}

    def augmentation(self, arrays, hyperparams, **kwargs):
        return arrays

    def apply_preprocess(self, xs, ys, row, key, params, **kwargs):
        """Applies additional feature preprocessing

        Currently supported:
            - (OUTPUT) Class loss weights
            - (INPUT) Coordinate localization map

        Args:
            xs (np.array): model array inputs
            ys (np.array): model array targets
            row (str): name of data sample (defined in client.db.header)
            key (str): name of specific input or output (defined in train.yml)
            params (dict): hyperparameter dict

        Returns:
            np.array or none: Returns new features if possible else none
        """

        if key == "coord":
            return positional_encoding.add_coordinates(xs, params, kwargs)
        elif key == "mask":
            return class_weights.modify_loss(xs, ys, row, params, kwargs)

        return

    def assign_preprocess(self, xs, ys, hyperparams, row, **kwargs):
        """Assigns preprocessing functions depending on train.yml config

        Args:
            xs (np.array): model array inputs
            ys (np.array): model array targets
            hyperparams (dict): hyperparameter dict
            row (str): name of data sample (defined in client.db.header)
        """

        for entry in hyperparams:
            params = hyperparams[entry]
            if params is not None:
                
                for key in params: 
                    if type(params[key]) == dict:
                        name = params[key]["name"]

                        if entry in ys:
                            feature = self.apply_preprocess(  # outputs
                                xs[name],
                                ys[entry],
                                row,
                                key,
                                params[key],
                                **kwargs,
                            )
                        else:
                            feature = self.apply_preprocess(  # inputs
                                xs[name],
                                ys[entry],
                                row,
                                key,
                                params[key],
                                **kwargs,
                            )

                        if feature is not None:
                            xs[name] = feature

    def preprocess(self, arrays, row, **kwargs):

        # --- Extract input / output fields
        inputs = self.hyperparams["train"]["xs"]
        outputs = self.hyperparams["train"]["ys"]

        # --- Additional custom data augmentation
        arrays = self.augmentation(arrays, self.hyperparams, **kwargs)

        # --- Get `ys` from `xs`
        xs = arrays["xs"]
        ys = {k: xs[k] for k in get_unique_subfields(outputs)}

        # --- Add additional features i.e. feature map coordinates
        self.assign_preprocess(xs, ys, inputs, row, **kwargs)

        # --- Set loss weights
        self.assign_preprocess(xs, ys, outputs, row, **kwargs)

        return arrays

    def create_generator(self, gen_data):
        for xs, ys in gen_data:
            yield xs, ys

    def create_generators(self, arr=None, test=False, **kwargs):
        if arr is None:
            gen_train, gen_valid = Client.create_generators(self, test=test, **kwargs)
            gen_train = self.create_generator(gen_train)
            gen_valid = self.create_generator(gen_valid)
            yield from (gen_train, gen_valid)

        else:
            while not test:
                for xs, ys in arr:
                    yield xs, ys

            for xs, ys in arr:
                yield xs, ys



"""Interface for getting available metrics"""

from tfcaidm.common.inherit import inherit
from tfcaidm.metrics.custom import _zoo


@inherit(_zoo.metrics)
def available_metrics():
    return custom_metric()


# --- metric in csv
def custom_metric():
    customs = {}
    return customs



"""UNet++ model"""

import math
import numpy as np

import tfcaidm.models.layers.extra as extra
import tfcaidm.models.nn.utils as utils
from tfcaidm.models.layers import transform


def skip_connections(x_list, ind_outer, ind_inner, ind_left):
    """Adds all intermediate skip connections"""

    x_skip = [x_list[ind_left]]

    for i in range(ind_outer):
        if ind_inner == i:
            break
        # --- Get next skip connection neuron
        ind_left -= ind_outer + 1 - i
        x_skip.append(x_list[ind_left])

    x_skip = transform.add(x_skip)

    return x_skip


def max_downsamples(x_dim, s_dim, n):
    """Determine amount of possible downsamples"""

    try:
        return int(math.log(x_dim, s_dim))
    except:
        return n


def unetpp(x, n, c, k, s, hyperparams):
    """Implementation of unet++ model"""

    # --- Control downsampling
    x_size = x.shape[1:-1]  # depth, width, height
    d_size = np.array(
        [
            max_downsamples(x_size[0], s[0], n),
            max_downsamples(x_size[1], s[1], n),
            max_downsamples(x_size[2], s[2], n),
        ]
    )

    # --- Refine input
    x = utils.in_conv(x, c, k, hyperparams)
    x = extra.layer_name(x, "eblock_0")
    print(x.shape)

    # --- Store intermediate layers and feature map channels
    x_list = [x]
    c_list = [c]
    e_list = [x]
    d_list = []
    f_list = []

    # --- Iteratively define unet
    for i in range(n):
        x, c = utils.down_conv(x, c, k, s, hyperparams)
        x = extra.layer_name(x, f"eblock_{i + 1}")

        e_list.append(x)
        x_list.append(x)
        c_list.append(c)
        bottom = -(i + 2)
        inner_size = d_size - 1
        print(x.shape)

        # --- Skip connections and upsampling
        for j in range(i + 1):
            x_skip = skip_connections(x_list, i, j, bottom)
            x, c = utils.up_conv(x, x_skip, c, k, s, inner_size, hyperparams)

            # --- Outer most decoder output
            if i == n - 1:
                x = extra.layer_name(x, f"dblock_{j}")
                d_list.append(x)

            # --- Full-sized feature map outputs
            if i == j:
                f_list.append(x)

            inner_size += 1
            x_list.append(x)
            c_list.append(c)
            print(x.shape)

        d_size -= 1
        x = x_list[bottom]
        c = c_list[bottom]

    return {
        "encoders": e_list,
        "decoders": d_list,
        "full_res": f_list,
    }



"""Pyramid scene parsing network block"""

import math

import tfcaidm.models.layers.transform as transform
from tfcaidm.models.utils import select as model_select


def pool_block(x, c, k, s, hyperparams):
    c = int(c / 2)

    # --- Average pool
    a = transform.average_pool(x, s)
    a = model_select.conv_selection(x=a, c=c, k=k, hyperparams=hyperparams)

    # --- Max pool
    m = transform.max_pool(x, s)
    m = model_select.conv_selection(x=m, c=c, k=k, hyperparams=hyperparams)

    return [a, m]


def upsample_block(x_, d, w, h):

    for i in range(len(x_)):

        # --- Control downsampling
        x_size = x_.shape[1:-1]  # depth, width, height
        d_ = max_downsamples(x_size[0], d, 1)
        w_ = max_downsamples(x_size[1], w, 1)
        h_ = max_downsamples(x_size[2], h, 1)

        # --- Get new feature dim
        depth = auto_stride(d, d_)
        width = auto_stride(w, w_)
        height = auto_stride(h, h_)
        s = (depth, width, height)

        # --- Upsample
        x_[i] = transform.up_sample(x_[i], s=s)

    return x_


def max_downsamples(x_dim, s_dim, n):
    """Determine amount of possible downsamples"""

    try:
        return int(math.log(x_dim, s_dim)) - n
    except:
        return 1


def auto_stride(a, b):
    return int(2 ** a / 2 ** b) if a >= b else 1


def psp(x, c, k, s, hyperparams, **kwargs):
    """Implementation of PSPNet block https://arxiv.org/abs/1612.01105"""

    n_branches = hyperparams["model"]["branches"]

    # --- Control downsampling
    x_size = x.shape[1:-1]  # depth, width, height
    d = max_downsamples(x_size[0], s[0], n_branches)
    w = max_downsamples(x_size[1], s[1], n_branches)
    h = max_downsamples(x_size[2], s[2], n_branches)

    x_list = [x]
    n_channels = int(c / n_branches) if c >= n_branches else 1

    # create parallel branches
    for i in range(n_branches):

        # --- Get new feature dim
        depth = auto_stride(d, i)
        width = auto_stride(w, i)
        height = auto_stride(h, i)
        s = (depth, width, height)

        # --- Process features
        x_ = pool_block(x, n_channels, k, s, hyperparams)
        x_ = upsample_block(x_, d, w, h)

        # --- Aggregate each upsampled feature
        for feature in x_:
            x_list.append(feature)

    # --- Combine feature maps
    x = transform.add(x_list)

    # --- Refine features
    x = model_select.conv_selection(x=x, c=c, k=1, hyperparams=hyperparams)

    return x



"""DenseNet block"""

import tfcaidm.models.layers.transform as transform
from tfcaidm.models.utils import select as model_select


def dense(x, c, k, hyperparams, **kwargs):
    """Implementation of dense-net block https://arxiv.org/abs/1608.06993"""

    # --- Temporary placement
    n = hyperparams["model"]["depth"]
    b = hyperparams["model"]["bneck"]
    o = c

    # --- Dense block
    x = transform.concat(
        [x, model_select.conv_selection(x=x, c=c, k=k, hyperparams=hyperparams)]
    )
    for i in range(1, n):
        x = dense_conv(x, i + 1, c, b, hyperparams)
        x = transform.concat(
            [x, model_select.conv_selection(x=x, c=c, k=k, hyperparams=hyperparams)]
        )

    # --- Transition layer
    x = transition_layer(x, c, o, k, hyperparams)

    return x


def dense_conv(x, n, c, b, hyperparams):
    """Dense conv to optionally compute bottleneck"""

    # --- Bottleneck if too many channels (num_layers > bottleneck ratio)
    if n >= b:
        x = model_select.conv_selection(x=x, c=(b * c), k=1, hyperparams=hyperparams)
    else:
        x = model_select.conv_selection(x=x, c=(n * c), k=1, hyperparams=hyperparams)

    return x


def transition_layer(x, c, o, k, hyperparams):
    """Transition layer following a dense block"""

    x = model_select.conv_selection(x=x, c=c, k=k, hyperparams=hyperparams)
    return model_select.conv_selection(x=x, c=o, k=1, hyperparams=hyperparams)



"""Tools to flatten and unflatten nested dicts"""

from tfcaidm.common.constants import DELIM


def unflatten_dict(flattened_dict, DELIM=DELIM):
    """Converts a flattened (non-nested) dict into a nested dict.

    Args:
        flattened_dict (dict): A flattened (non-nested) dict

    Returns:
        [dict]: Nested-dict.

    Example:

    flattened_dict = {
        "train/xs/dat": None,
        "train/ys/pna/mask_id": ["msk-pna"],
        "train/ys/pna/mask_weight": [1, 100],
        "train/ys/pna/metrics": ["dice"],
        "train/ys/ratio/mask_id": ["msk-ratio"],
        "train/ys/ratio/mask_weight": [0, 5],
        "train/ys/ratio/metrics": ["mse"],
        "train/trainer/seed": [0],
        "train/trainer/n_folds": [1],
        "train/trainer/batch_size": [8],
        "train/trainer/iters": [10000],
        "train/trainer/steps": [100],
        "train/trainer/valid_freq": [4],
        "train/trainer/lr": ["6e-3"],
        "train/trainer/lr_alpha": [0.25],
        "train/trainer/lr_decay": [0.97],
    }

    nested_dict = unflatten_dict(flattened_dict)

    ---

    nested_dict = {
        "train": {
            "xs": {"dat": None},
            "ys": {
                "pna": {
                    "mask_id": ["msk-pna"],
                    "mask_weight": [1, 100],
                    "metrics": ["dice"],
                },
                "ratio": {
                    "mask_id": ["msk-ratio"],
                    "mask_weight": [0, 5],
                    "metrics": ["mse"],
                },
            },
            "trainer": {
                "seed": [0],
                "n_folds": [1],
                "batch_size": [8],
                "iters": [10000],
                "steps": [100],
                "valid_freq": [4],
                "lr": ["6e-3"],
                "lr_alpha": [0.25],
                "lr_decay": [0.97],
            },
        }
    }

    """

    def insert(dct, lst):
        for x in lst[:-2]:
            dct = dct.setdefault(x, {})
        dct[lst[-2]] = lst[-1]

    nested_dict = {}

    lists = ([*k.split(DELIM), v] for k, v in flattened_dict.items())

    for lst in lists:
        insert(nested_dict, lst)

    return nested_dict


def flatten_dict(nested_dict, flattened_dict=None, parent_key=None, DELIM=DELIM):
    """Convert a nested dict into a flattened dict.

    Args:
        nested_dict (dict): A nested dict.
        flattened_dict (dict): Mutated during recursion, value should remain None.
        parent_key (str): Parent's key name, value should remain None.

    Returns:
        [dict]: Flattened dict.

    Example:

    nested_dict = {
        "train": {
            "xs": {"dat": None},
            "ys": {
                "pna": {
                    "mask_id": ["msk-pna"],
                    "mask_weight": [1, 100],
                    "metrics": ["dice"],
                },
                "ratio": {
                    "mask_id": ["msk-ratio"],
                    "mask_weight": [0, 5],
                    "metrics": ["mse"],
                },
            },
            "trainer": {
                "seed": [0],
                "n_folds": [1],
                "batch_size": [8],
                "iters": [10000],
                "steps": [100],
                "valid_freq": [4],
                "lr": ["6e-3"],
                "lr_alpha": [0.25],
                "lr_decay": [0.97],
            },
        }
    }

    flattened_dict = flatten_dict(nested_dict)

    ---

    flattened_dict = {
        "train/xs/dat": None,
        "train/ys/pna/mask_id": ["msk-pna"],
        "train/ys/pna/mask_weight": [1, 100],
        "train/ys/pna/metrics": ["dice"],
        "train/ys/ratio/mask_id": ["msk-ratio"],
        "train/ys/ratio/mask_weight": [0, 5],
        "train/ys/ratio/metrics": ["mse"],
        "train/trainer/seed": [0],
        "train/trainer/n_folds": [1],
        "train/trainer/batch_size": [8],
        "train/trainer/iters": [10000],
        "train/trainer/steps": [100],
        "train/trainer/valid_freq": [4],
        "train/trainer/lr": ["6e-3"],
        "train/trainer/lr_alpha": [0.25],
        "train/trainer/lr_decay": [0.97],
    }

    """

    if flattened_dict is None:
        flattened_dict = {}

    for k, v in nested_dict.items():
        k = f"{parent_key}{DELIM}{k}" if parent_key else k
        if isinstance(v, dict):
            flatten_dict(nested_dict=v, flattened_dict=flattened_dict, parent_key=k)
            continue

        flattened_dict[k] = v

    return flattened_dict



"""Defined model blocks"""

import tfcaidm.models.layers.learner as learner
import tfcaidm.models.layers.transform as transform

import tfcaidm.models.nn.ae as ae
import tfcaidm.models.nn.unet as unet
import tfcaidm.models.nn.unetpp as unetpp
import tfcaidm.models.nn.unet3p as unet3p

import tfcaidm.models.blocks.atrous_block as atrous_block
import tfcaidm.models.blocks.attention_gate as attention_gate
import tfcaidm.models.blocks.cbam_block as cbam_block
import tfcaidm.models.blocks.csp_block as csp_block
import tfcaidm.models.blocks.convgru_cell as convgru_cell
import tfcaidm.models.blocks.dense_block as dense_block
import tfcaidm.models.blocks.eca_block as eca_block
import tfcaidm.models.blocks.inception_block as inception_block
import tfcaidm.models.blocks.psp_block as psp_block
import tfcaidm.models.blocks.se_block as se_block
import tfcaidm.models.blocks.separable_conv as separable_conv
import tfcaidm.models.blocks.u2net_block as u2net_block

import tfcaidm.models.utils.head as head
import tfcaidm.models.utils.tasks as tasks

conv_types = {
    "conv": learner.conv,
    "small": separable_conv.separable_conv,
    "smaller": separable_conv.smaller_separable_conv,
    "smallest": separable_conv.smallest_separable_conv,
    "channel_spatial": separable_conv.channel_spatial_conv,
}

tran_types = {
    "conv": learner.tran,
    "small": separable_conv.separable_tran,
    "smaller": separable_conv.smaller_separable_tran,
    "smallest": separable_conv.smallest_separable_tran,
    "channel_spatial": separable_conv.channel_spatial_tran,
}

pool_types = {
    "conv": learner.conv,
    "max": transform.max_pool,
    "avg": transform.average_pool,
    "aspp": atrous_block.aspp,
    "acsp": atrous_block.acsp,
    "wasp": atrous_block.wasp,
}

eblocks = {
    "conv": learner.conv,
    "aspp": atrous_block.aspp,
    "acsp": atrous_block.acsp,
    "wasp": atrous_block.wasp,
    "cbam": cbam_block.cbam,
    "csp": csp_block.csp,
    "dense": dense_block.dense,
    "eca": eca_block.eca,
    "inception": inception_block.inception,
    "psp": psp_block.psp,
    "se": se_block.se,
    "u2net": u2net_block.u2net,
}

dblocks = {
    "conv": learner.tran,
    "convgru": convgru_cell.convgru,
    "attention": attention_gate.attention_gate,
}

heads = {
    "encoder_classifier": head.Encoder.last_layer,
    "encoder_multi_scale_classifier": head.Encoder.multi_scale,
    "decoder_classifier": head.Decoder.last_layer,
    "decoder_multi_scale_classifier": head.Decoder.multi_scale,
    "decoder_deep_supervision_classifier": head.Decoder.deep_supervision,
    "decoder_complex_supervision_classifier": head.Decoder.complex_supervision,
}

task_types = {
    "auto": tasks.auto_task_selector,
}

models = {
    "ae": ae.ae,
    "unet": unet.unet,
    "unet++": unetpp.unetpp,
    "unet3+": unet3p.unet3p,
}



"""Format and store hyperparameters dict"""

import ast

from tfcaidm.common.constants import DELIM
from tfcaidm.jobs.utils.tool import flatten_dict, unflatten_dict


class HyperParameters:
    def __init__(self, hyperparams, *args, **kwargs):
        hyperparams = HyperParameters.read(hyperparams)
        self.__hyperparams = HyperParameters.unflatten(hyperparams)
        self.results = HyperParameters.unflatten(hyperparams)

    @property
    def hyperparams(self):
        return self.__hyperparams

    @staticmethod
    def read(hyperparams):
        params = {}

        for k, v in hyperparams.items():
            try:
                v = ast.literal_eval(v)
            except:
                pass
            params[k] = v

        return params

    @staticmethod
    def flatten(hyperparams):
        return flatten_dict(hyperparams, DELIM=DELIM)

    @staticmethod
    def unflatten(hyperparams):
        return unflatten_dict(hyperparams, DELIM=DELIM)



"""Setup python environment"""


import os
import pytz
import getpass
import inspect
from datetime import datetime


def get_pst_time():
    """Function to get PDT time and date in 12hr format"""
    date_format = "%m_%d_%Y_%H_%M_%S_%Z"
    date_format = "%b %d %Y %-I:%M %p"
    date = datetime.now(tz=pytz.utc)
    date = date.astimezone(pytz.timezone("US/Pacific"))
    pstDateTime = date.strftime(date_format)
    return pstDateTime


def setup(user, param_path, row_id, python_path, log_dir, **kwargs):
    """Function to setup environment

    Args:
        user        (str): account username
        param_path  (str): hyper-parameter csv path
        row_id      (int): rid in the param_path
        python_path (str): python path
        log_dir     (str): experiment output directory
    """

    env = f"""
        echo "{get_pst_time()}" > {log_dir}/stdout

        export JARVIS_PATH_CONFIGS=/home/{user}/.jarvis
        export JARVIS_PARAMS_CSV={param_path}
        export JARVIS_PARAMS_ROW={row_id}
        export PYTHONPATH={python_path}
    """

    return inspect.cleandoc(env)


def install(log_dir, libraries, **kwargs):
    """Function to pip install needed requirements

    Args:
        libraries (list): list of tuples containing library and version
        log_dir   (str): experiment output directory
    """

    inst = ""

    for library in libraries:
        lib, ver = library
        inst += f"{verify(lib, ver)} >> {log_dir}/stdout"
        inst += "\n"

    return inspect.cleandoc(inst)


def verify(library, version, **kwargs):
    """Function to check if install already exists

    Args:
        library (str): name of library to install
        version (str): version of library
    """

    ver = f"""
        ver=$(pip list | grep {library} | head -1 | xargs)
        if [ "$ver" != "{library} {version}" ]
        then
            pip install {library}=={version}
        fi
    """

    return ver


def run(program, log_dir, **kwargs):
    """Function to run python file

    Args:
        program (str): program filename to run
        log_dir (str): experiment output directory
    """

    exe = f"""
        python {program} >> {log_dir}/stdout 2>&1
    """

    return inspect.cleandoc(exe)


def set_requirements(**kwargs):
    """Function to setup all requirements and environment variables"""

    setup_sequence = []
    setup_sequence.append(setup(**kwargs))
    setup_sequence.append(install(**kwargs))
    setup_sequence.append(run(**kwargs))

    command = ""

    for step in setup_sequence:
        command += step
        command += "\n"

    return command


if __name__ == "__main__":
    # --- Example arguments
    kwargs = {}

    kwargs["user"] = getpass.getuser()
    kwargs["param_path"] = os.getcwd()
    kwargs["python_path"] = os.getcwd()
    kwargs["log_dir"] = os.getcwd()
    kwargs["program"] = os.getcwd() + "/" + "main.py"
    kwargs["row_id"] = 0

    kwargs["libraries"] = [
        ("tensorflow", "2.5.0"),
        ("pytorch", "1.9.0"),
        ("torchvision", "0.10.0"),
        ("pytorch-lightning", "1.4.2"),
    ]

    command = set_requirements(**kwargs)
    print(command)



"""Distance related loss functions"""

from tensorflow import losses
from tensorflow.keras import layers


class MeanAbsoluteError(layers.Layer):
    """Computes the mean absolute error (L1) loss"""

    def __init__(self, name="MeanAbsoluteError", **kwargs):
        super(MeanAbsoluteError, self).__init__(name=name, **kwargs)
        self.loss = losses.MeanAbsoluteError()

    def call(self, y_true, y_pred, weights=None, add_loss=True, **kwargs):
        loss = self.loss(
            y_true=y_true,
            y_pred=y_pred,
            sample_weight=weights,
        )

        if add_loss:
            self.add_loss(loss)

        return loss


class MeanSquaredError(layers.Layer):
    """Computes the mean squared error (L2) loss"""

    def __init__(self, name="MeanSquaredError", **kwargs):
        super(MeanSquaredError, self).__init__(name=name, **kwargs)
        self.loss = losses.MeanSquaredError()

    def call(self, y_true, y_pred, weights=None, add_loss=True, **kwargs):
        loss = self.loss(
            y_true=y_true,
            y_pred=y_pred,
            sample_weight=weights,
        )

        if add_loss:
            self.add_loss(loss)

        return loss



"""Sets random seeds"""

import os
import random
import numpy as np
import tensorflow as tf


def set_seeds(seed=0):
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    tf.random.set_seed(seed)


def set_determinism(
    seed=0,
    fast=False,
    single_thread=False,
    *args,
    **kwargs,
):
    """
    Enable 100% reproducibility on operations related to tensor and randomness.

    ref: https://suneeta-mall.github.io/2019/12/22/Reproducible-ml-tensorflow.html

    Args:
        seed (int): seed value for global randomness
        fast (bool): faster performance at the cost of determinism/reproducibility
        single_thread (bool): disable thread parallelism between independent operations
    """
    set_seeds(seed=seed)

    if fast:
        return

    os.environ["TF_DETERMINISTIC_OPS"] = "1"
    os.environ["TF_CUDNN_DETERMINISTIC"] = "1"

    if single_thread:
        tf.config.threading.set_inter_op_parallelism_threads(1)
        tf.config.threading.set_intra_op_parallelism_threads(1)



"""Normalization layers"""

import tensorflow as tf
from tensorflow.keras import layers, regularizers


normalization_kwargs = {
    "bnorm": layers.BatchNormalization,
    "lnorm": layers.LayerNormalization,
    "none": lambda: tf.identity,
}


def normalization(hyperparams):
    k = hyperparams["model"]["norm"]

    if k not in normalization_kwargs:
        raise ValueError(f"ERROR! Normalization layer `{k}` is not defined!")

    x = normalization_kwargs[k]

    return x


dropout = lambda x, r=0.25: layers.Dropout(rate=r)(x)
l1 = lambda l1_rate=0.1: regularizers.L1(l1=l1_rate)
l2 = lambda l2_rate=0.1: regularizers.L2(l2=l2_rate)
l1_l2 = lambda l1_rate=0.1, l2_rate=0.1: regularizers.l1_l2(l1=l1_rate, l2=l2_rate)



"""Automatically determine output shape for different model tasks"""

import tfcaidm.common.constants as constants
import tfcaidm.models.layers.learner as learner
import tfcaidm.models.layers.transform as transform
import tfcaidm.models.layers.activation as activation

from tfcaidm.losses.loss import Loss
from tfcaidm.metrics.metric import Metric
from tfcaidm.models.utils import select as model_select


def auto_task_selector(x, inputs, client):
    logits = {}

    hyperparams = client.hyperparams
    outputs = hyperparams["train"]["ys"]
    output_shapes = client.get_output_shapes()

    for name, addons in outputs.items():
        feature = model_select.head_selection(x, addons)
        logit_name = constants.get_name(name, "logits")

        if len(output_shapes[name]) == 1:
            logits[logit_name] = scaler_projection(
                x=feature,
                n_classes=addons["n_classes"],
                hyperparams=hyperparams,
                name=logit_name,
            )

        else:
            logits[logit_name] = vector_projection(
                x=feature,
                y=output_shapes[name],
                n_classes=addons["n_classes"],
                hyperparams=hyperparams,
                name=logit_name,
            )

        # --- Get inputs and outputs
        y_true = inputs[name]
        y_pred = logits[logit_name]

        # --- Get loss weights and alpha
        weights = get_weights(inputs, addons)
        alpha = hyperparams["train"]["trainer"]["lr_alpha"]

        # --- Get names
        loss_name = hyperparams["train"]["ys"][name]["loss"]
        metric_name = hyperparams["train"]["ys"][name]["metric"]

        # --- Get loss and metric
        loss = Loss.add_loss(
            y_true, y_pred, name, loss_name, alpha=alpha, weights=weights
        )
        metric = Metric.add_metric(
            y_true, y_pred, name, metric_name, alpha=alpha, weights=weights
        )

        logits.update(loss)
        logits.update(metric)

    return logits


def get_weights(inputs, addons, _id="mask_id"):
    weights = None

    for k in addons:
        if type(addons[k]) == dict:
            if weights is None:
                weights = inputs[addons[k]["name"]]
            else:
                weights += inputs[addons[k]["name"]]

    return weights


# --- Helpers


def scaler_projection(x, n_classes, hyperparams, name=None):
    x = activation.attn_msk(x, hyperparams)
    x = transform.global_average_pool(x)

    return learner.pre_activation_mlp(x=x, c=n_classes, name=name)


def vector_projection(x, y, n_classes, hyperparams, name=None):
    """Reshape for [depth, width, height]"""

    # --- Get shape of each dimension
    xd, xw, xh = x.shape[-4:-1]
    yd, yw, yh = y[:-1]

    # --- Reshape and linearly project
    s1 = reshape_dim(xd, yd)
    s2 = reshape_dim(xw, yw)
    s3 = reshape_dim(xh, yh)

    return learner.pre_activation_conv(x=x, c=n_classes, k=1, s=(s1, s2, s3), name=name)


def reshape_dim(x, y):

    # --- Identity
    if x == y:
        return int(x / y)

    # --- Downsample
    elif x > y:
        return int(x / y)

    # --- Upsample
    elif x < y:
        return int(y / x)



"""Model.fit compatible callbacks"""

import os
from tensorflow.keras import callbacks


def lr_scheduler(hyperparams):
    lr_decay = hyperparams["train"]["trainer"]["lr_decay"]
    assert lr_decay >= 0 and lr_decay <= 1

    return callbacks.LearningRateScheduler(
        lambda epoch, lr: lr * lr_decay if epoch > 10 else lr
    )


def model_checkpoints(hyperparams):
    # --- Create output_dir
    log_dir = hyperparams["train"]["trainer"]["log_dir"] + "/weights/"
    os.makedirs(log_dir, exist_ok=True)

    # --- Create weight file
    path = hyperparams["model"]["model"] + "_" + "{epoch:03d}.hdf5"
    filepath = log_dir + path

    return callbacks.ModelCheckpoint(
        filepath,
        monitor="val_loss",
        verbose=True,
        save_best_only=True,
        save_weights_only=True,
        mode="min",
        save_freq="epoch",
    )


def tensorboard_init(hyperparams):
    # --- Create output_dir
    log_dir = hyperparams["train"]["trainer"]["log_dir"] + "/logdirs/"
    return callbacks.TensorBoard(log_dir=log_dir)


def exit_on_nan(hyperparams):
    return callbacks.TerminateOnNaN()


# --- NOTE: Not implemented yet
def reduce_lr(hyperparams, **kwargs):
    callbacks.ReduceLROnPlateau(
        monitor="val_loss",
        factor=0.1,
        patience=10,
        verbose=0,
        mode="auto",
        min_delta=0.0001,
        cooldown=0,
        min_lr=0,
        **kwargs
    )


def early_stop(hyperparams, **kwargs):
    return callbacks.EarlyStopping(
        monitor="val_loss",
        min_delta=0,
        patience=0,
        verbose=0,
        mode="auto",
        baseline=None,
        restore_best_weights=False,
    )



"""Summarizes training results"""

import fcntl
import pandas as pd
from pathlib import Path


VAL = "val"


class Summary:
    def __init__(self, histories):
        """Summarize training run.

        Args:
            histories ([type]): model.fit history.history dict.
        """

        if type(histories) is list:
            self.histories = histories
        else:
            self.histories = [histories]

    def summarize_eval(self):
        folds = {"fold": {}}

        for i, history in enumerate(self.histories):
            folds["fold"][i] = history

        average = to_dataframe(folds["fold"]).mean(axis=1)

        folds["fold"]["avg"] = {}
        folds["fold"]["avg"]["last"] = {k: v for k, v in average.items()}

        return folds

    def summarize_best_history(self, valid_freq, name="val_loss"):
        assert name is None or VAL in name, f"param name must contain `{VAL}`"
        assert valid_freq > 0, f"valid_freq={valid_freq} but must be > 0"

        folds = {"fold": {}}

        for i, history in enumerate(self.histories):
            df = to_dataframe(history)
            folds["fold"][i] = self.best(df, valid_freq, name)

        average = to_dataframe(folds["fold"]).mean(axis=1)

        folds["fold"]["avg"] = {}
        folds["fold"]["avg"]["best"] = {k: v for k, v in average.items()}

        return folds

    def summarize_last_history(self, valid_freq):
        assert valid_freq > 0, f"valid_freq={valid_freq} but must be > 0"

        folds = {"fold": {}}

        for i, history in enumerate(self.histories):
            df = to_dataframe(history)
            folds["fold"][i] = self.last(df)

        average = to_dataframe(folds["fold"]).mean(axis=1)

        folds["fold"]["avg"] = {}
        folds["fold"]["avg"]["last"] = {k: v for k, v in average.items()}

        return folds

    def per_fold(self, history, valid_freq, name="val_loss"):
        df = to_dataframe(history)
        last = self.last(df)
        best = self.best(df, valid_freq, name)

        return {"best": best, "last": last}

    def best(self, df, valid_freq, name):
        if "loss" in name:
            best_ind = get_min_index(df, name=name)
        else:
            best_ind = get_max_index(df, name=name)

        adjusted_ind = valid_freq * (best_ind + 1) - 1

        best = {}
        best["epoch"] = adjusted_ind
        best.update(
            {k: (df[k][best_ind] if VAL in k else df[k][adjusted_ind]) for k in df}
        )

        return best

    def last(self, df):
        last_ind = {k: df[k].last_valid_index() for k in df}

        last = {}
        last["epoch"] = last_ind["loss"]
        last.update({k: df[k][last_ind[k]] for k in df})

        return last


def get_max_index(df, name=None):
    if name is None or name not in df:
        return df.idxmax(axis=0, skipna=True)
    return df.idxmax(axis=0, skipna=True)[name]


def get_min_index(df, name=None):
    if name is None or name not in df:
        return df.idxmin(axis=0, skipna=True)
    return df.idxmin(axis=0, skipna=True)[name]


def to_dataframe(results):
    return pd.DataFrame.from_dict(results, orient="index").transpose()


def thread_safe_csv_writes(df, path, mode, header):
    with open(path, mode) as f:
        fcntl.flock(f, fcntl.LOCK_EX)
        df.to_csv(f, mode=mode, header=header, index_label="run_id", index=False)
        fcntl.flock(f, fcntl.LOCK_UN)


def save_results(hyperparams):
    df = pd.DataFrame([hyperparams])
    path = hyperparams["env/path/param_csv"].replace("hyper.csv", "results.csv")
    csv_path = Path(path)

    # --- Save csv
    if not csv_path.is_file():
        thread_safe_csv_writes(df, path, mode="w", header=True)
    else:
        thread_safe_csv_writes(df, path, mode="a", header=False)

    print(f"- csv saved to {path}")



"""Handle all loss related functions"""

import tfcaidm.common.constants as constants
import tfcaidm.losses.custom.registry as registry


class Loss:
    def __init__(self):
        pass

    def create(self):
        pass

    @classmethod
    def add_loss(
        self, y_true, y_pred, output_name, loss_name, *, alpha=0.5, weights=None
    ):
        """Add custom losses to model"""

        name = constants.get_name(output_name, "loss")
        name = constants.get_name(name, loss_name)

        loss = self.loss_selection(
            self,
            name=name,
            func=loss_name,
            weights=weights,
            alpha=alpha,
        )(y_true=y_true, y_pred=y_pred)

        return {name: loss}

    def loss_selection(
        self,
        name,
        func,
        weights=1.0,
        alpha=0.75,
        gamma=0.75,
        class_of_interest=1,
        **kwargs,
    ):
        """Selects loss function to use"""

        def zoo(y_true, y_pred):
            """Utility to choose between different loss functions"""

            loss_fns = registry.available_losses()

            if func not in loss_fns:
                raise ValueError(f"ERROR! Loss function `{func}` is not defined!")

            loss_fn = loss_fns[func](name=name)

            return loss_fn(
                y_true=y_true,
                y_pred=y_pred,
                weights=weights,
                alpha=alpha,
                gamma=gamma,
                class_of_interest=class_of_interest,
            )

        def loss(y_true, y_pred):
            return zoo(y_true, y_pred)

        return loss



"""Selects training callbacks to use"""

import tfcaidm.train.custom.registry as registry

# --- callbacks in csv
def callback_selection(hyperparams):
    """get training callback functions"""

    callbacks = registry.available_callbacks()

    # --- Extract hyperparams from params csv
    cbacks = hyperparams["train"]["trainer"]["callbacks"]

    return [callbacks[k] for k in cbacks if k in callbacks]



"""Convolutional gated recurrent cell"""

import tfcaidm.models.layers.activation as activation
import tfcaidm.models.layers.learner as learner
import tfcaidm.models.layers.transform as transform


def project(x, c):
    if x.shape[-1] != c:
        return learner.conv_layer(c, 1)(x)
    return x


def convgru(x, x_skip, c, k, hyperparams, s=(1, 2, 2), **kwargs):
    """Implementation of conv-gru cell https://arxiv.org/abs/1412.3555"""

    # --- Project channels
    x = project(x, c)
    x_skip = project(x_skip, c)

    wz = learner.conv_layer(c, k)  # update gate
    wr = learner.conv_layer(c, k)  # reset gate
    wh = learner.conv_layer(c, k)  # gate fusion

    pool = lambda x, s: (transform.max_pool(x, s) + transform.average_pool(x, s)) / 2
    layer = lambda x, hyperparams: activation.apply_all_activations(x, hyperparams)

    x_skip = pool(x_skip, s)

    z = layer(wz(x), hyperparams) + layer(wz(x_skip), hyperparams)
    r = layer(wr(x), hyperparams) + layer(wr(x_skip), hyperparams)
    h_ = activation.tanh(wh(x) + wh(x_skip) * r)
    h = (1 - z) * x_skip + z * h_

    # --- Upsample
    if s != (1, 1, 1):
        h = transform.up_sample(x=h, s=s)

    return h



"""Efficient channel attention block"""

import math

import tfcaidm.models.layers.activation as activation
import tfcaidm.models.layers.learner as learner
import tfcaidm.models.layers.transform as transform


def adaptive_kernel_size(c, scale=2, shift=1):
    """
    High-dimensional channels = longer range interactions
    Low-dimensional channels = shorter range interactions
    """

    t = int(abs((math.log(c, 2) + shift) / scale))
    return t if t % 2 else t + 1


def eca(x, c, hyperparams, **kwargs):
    """Implementation of efficient channel attention (ECA) block https://arxiv.org/abs/1910.03151"""

    # --- Squeeze (pool and combine feature maps)
    gmp = transform.global_max_pool(x, keepdims=True)
    gap = transform.global_average_pool(x, keepdims=True)

    # --- Adaptive kernel size
    c = gmp.shape[-1]
    k = (1, 1, adaptive_kernel_size(c))

    # --- Excite (convolution over squeezed features)
    conv = learner.conv_layer(c=c, k=k)
    gpl = conv(gmp) + conv(gap)

    # --- Attention
    scale = activation.attn_msk(gpl, hyperparams)

    return x * scale



"""Handle all metric related functions"""

import tfcaidm.common.constants as constants
import tfcaidm.metrics.custom.registry as registry


class Metric:
    def __init__(self):
        pass

    def create(self):
        pass

    @classmethod
    def add_metric(
        self, y_true, y_pred, output_name, metric_name, *, alpha=0.5, weights=None
    ):
        """Add custom metrics to model"""

        name = constants.get_name(output_name, "metric")
        name = constants.get_name(name, metric_name)

        metric = self.metric_selection(
            self,
            name=name,
            func=metric_name,
            weights=weights,
            alpha=alpha,
        )(y_true=y_true, y_pred=y_pred)

        return {name: metric}

    def metric_selection(
        self,
        name,
        func,
        weights=None,
        class_of_interest=1,
        **kwargs,
    ):
        """Selects metric function to use"""

        def zoo(y_true, y_pred):
            """Utility to choose between different metric functions"""

            metrics = registry.available_metrics()

            if func not in metrics:
                raise ValueError(f"ERROR! Metric function `{func}` is not defined!")

            metric = metrics[func](name=name)

            return metric(
                y_true=y_true,
                y_pred=y_pred,
                weights=weights,
                class_of_interest=class_of_interest,
            )

        def metric(y_true, y_pred):
            return zoo(y_true, y_pred)

        return metric



"""Model evaluation interface"""

import numpy as np
from scipy import stats
from scipy.spatial import distance
from sklearn import metrics


class Base:
    def __init__(self, model):
        self.model = model

    def forward(self, x):
        k = self.model.output_names
        v = self.model.predict(x)
        outputs = {k[i]: v[k[i]] for i in range(len(k))}

        return outputs

    def sort(self, x, best="high"):
        if best == "high":
            i = np.argsort(-x)
        elif best == "low":
            i = np.argsort(x)
        else:
            raise ValueError("arg `best` must be either 'high' or 'low'")

        return i

    @staticmethod
    def flatten(func):
        def apply(self, y_true, y_pred, *args, **kwargs):
            return func(self, y_true.flatten(), y_pred.flatten(), *args, **kwargs)

        return apply


class Classifier(Base):
    def __init__(self, model):
        super(Classifier, self).__init__(model)

    @Base.flatten
    def balanced_accuracy_score(
        self, y_true, y_pred, *, sample_weight=None, adjusted=False
    ):
        """Same as sklearn.metrics.balanced_accuracy_score"""

        return metrics.balanced_accuracy_score(
            y_true, y_pred, sample_weight=sample_weight, adjusted=adjusted
        )

    @Base.flatten
    def classification_report(
        self,
        y_true,
        y_pred,
        *,
        labels=None,
        target_names=None,
        sample_weight=None,
        digits=2,
        output_dict=False,
        zero_division="warn",
    ):
        """Same as sklearn.metrics.classification_report"""

        return metrics.classification_report(
            y_true,
            y_pred,
            labels=labels,
            target_names=target_names,
            sample_weight=sample_weight,
            digits=digits,
            output_dict=output_dict,
            zero_division=zero_division,
        )

    @Base.flatten
    def confusion_matrix(
        self, y_true, y_pred, *, labels=None, sample_weight=None, normalize=None
    ):
        """Same as sklearn.metrics.confusion_matrix"""

        return metrics.confusion_matrix(
            y_true,
            y_pred,
            labels=labels,
            sample_weight=sample_weight,
            normalize=normalize,
        )

    @Base.flatten
    def jaccard_score(
        self,
        y_true,
        y_pred,
        *,
        labels=None,
        pos_label=1,
        average="binary",
        sample_weight=None,
        zero_division="warn",
    ):
        """Same as sklearn.metrics.jaccard_score"""

        return metrics.jaccard_score(
            y_true,
            y_pred,
            labels=labels,
            pos_label=pos_label,
            average=average,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )

    @Base.flatten
    def f1_score(
        self,
        y_true,
        y_pred,
        *,
        labels=None,
        pos_label=1,
        average="binary",
        sample_weight=None,
        zero_division="warn",
    ):
        """Same as sklearn.metrics.f1_score"""

        return metrics.f1_score(
            y_true,
            y_pred,
            labels=labels,
            pos_label=pos_label,
            average=average,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )

    @Base.flatten
    def matthews_corrcoef(self, y_true, y_pred, *, sample_weight=None):
        """Same as sklearn.metrics.matthews_corrcoef"""

        return metrics.matthews_corrcoef(y_true, y_pred, sample_weight=sample_weight)


class Regressor(Base):
    def __init__(self, model):
        super(Regressor, self).__init__(model)

    def cdist(self, XA, XB, metric="euclidean", *, out=None, **kwargs):

        """Same as scipy.spatial.distance.cdist"""

        return distance.cdist(XA, XB, metric=metric, out=out, **kwargs)

    @Base.flatten
    def mean_absolute_error(
        self, y_true, y_pred, *, sample_weight=None, multioutput="uniform_average"
    ):
        """Same as sklearn.metrics.mean_absolute_error"""

        return metrics.mean_absolute_error(
            y_true, y_pred, sample_weight=sample_weight, multioutput=multioutput
        )

    @Base.flatten
    def mean_squared_error(
        self,
        y_true,
        y_pred,
        *,
        sample_weight=None,
        multioutput="uniform_average",
        squared=True,
    ):
        """Same as sklearn.metrics.mean_squared_error"""

        return metrics.mean_squared_error(
            y_true,
            y_pred,
            sample_weight=sample_weight,
            multioutput=multioutput,
            squared=squared,
        )

    @Base.flatten
    def pearsonr(self, y_true, y_pred):
        """Same as scipy.stats.pearsonr"""

        return stats.pearsonr(y_true, y_pred)

    @Base.flatten
    def r2_score(
        self, y_true, y_pred, *, sample_weight=None, multioutput="uniform_average"
    ):
        """Same as sklearn.metrics.r2_score"""

        return metrics.r2(
            y_true, y_pred, sample_weight=sample_weight, multioutput=multioutput
        )



"""Focal related loss functions"""

import tensorflow as tf

from tfcaidm.losses.funcs.entropy import WeightedCategoricalCrossentropy

reduce = lambda x: tf.math.reduce_sum(x)


class FocalLoss(WeightedCategoricalCrossentropy):
    """Implementation of binary focal loss"""

    def __init__(self, name="FocalLoss", **kwargs):
        super(FocalLoss, self).__init__(name=name, **kwargs)

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        gamma=2.0,
        class_of_interest=1,
        epsilon=1e-6,
        add_loss=True,
        **kwargs,
    ):
        # --- Weighted cross-entropy loss (log-term)
        loss = WeightedCategoricalCrossentropy.call(
            self,
            y_true=y_true,
            y_pred=y_pred,
            weights=weights,
            class_of_interest=class_of_interest,
            add_loss=False,
        )

        # ---- Extract the class of interest
        y_true = tf.cast(y_true[..., 0] == class_of_interest, tf.float32)
        y_pred = tf.nn.softmax(y_pred, axis=-1)[..., class_of_interest]

        # --- Calculate modulation to pos and neg labels
        modulation_pos = (1 - y_pred) ** gamma
        modulation_neg = y_pred ** gamma

        mask = tf.dtypes.cast(y_true, dtype=tf.bool)
        modulation = tf.where(mask, modulation_pos, modulation_neg)

        loss = reduce(modulation * loss)

        if add_loss:
            self.add_loss(loss)

        return loss



"""(EXPERIMENTAL) Depthwise separable convolution blocks https://arxiv.org/abs/1610.02357v3"""

import tensorflow as tf
import tfcaidm.models.layers.learner as learner
import tfcaidm.models.layers.transform as transform


# --- Alternative to depthwise operating on channels first then spatials
def channel_spatial_conv(x, c, k, s, d, name, hyperparams):
    channel = learner.conv(x=x, c=1, k=1, s=1, d=d, hyperparams=hyperparams)
    spatial = learner.conv(x=channel, c=c, s=s, d=d, hyperparams=hyperparams)
    return spatial


def channel_spatial_tran(x, c, k, s, d, name, hyperparams):
    channel = learner.conv(x=x, c=1, k=1, s=1, d=d, hyperparams=hyperparams)
    spatial = learner.tran(x=channel, c=c, s=s, d=d, hyperparams=hyperparams)
    return spatial


# --- Sub-modules
def depthwise_operation(x, k, s, d, hyperparams):
    x_list = []

    for i in range(x.shape[-1]):
        fm = tf.expand_dims(x[..., i], axis=-1)
        x_list.append(learner.conv(x=fm, c=1, k=k, s=s, d=d, hyperparams=hyperparams))

    return x_list


def depthwise_upsample_operation(x, k, s, d, hyperparams):
    x_list = []

    for i in range(x.shape[-1]):
        fm = tf.expand_dims(x[..., i], axis=-1)
        x_list.append(learner.tran(x=fm, c=1, k=k, s=s, d=d, hyperparams=hyperparams))

    return x_list


def extreme_reduction(x, hyperparams):
    x = learner.conv(x=x, c=1, k=1, hyperparams=hyperparams)
    return x


def pointwise_operation(x, c, hyperparams):
    x = learner.conv(x=x, c=c, k=1, hyperparams=hyperparams)
    return x


# --- Complete-modules
def separable_conv(x, c, k, s, d, name, hyperparams):
    # --- Depthwise Operation
    x = depthwise_operation(x, k, s, d, hyperparams)
    x = transform.concat(x)

    # --- Pointwise Operation
    x = pointwise_operation(x, c, hyperparams)
    return x


def separable_tran(x, c, k, s, d, name, hyperparams):
    # --- Depthwise Operation
    x = depthwise_upsample_operation(x, k, s, d, hyperparams)
    x = transform.concat(x)

    # --- Pointwise Operation
    x = pointwise_operation(x, c, hyperparams)
    return x


def smaller_separable_conv(x, c, k, s, d, name, hyperparams):
    # --- Extreme Reduction
    x = extreme_reduction(x, hyperparams)

    # --- Depthwise Operation
    x = depthwise_operation(x, k, s, d, hyperparams)
    x = transform.concat(x)

    # --- Pointwise Operation
    x = pointwise_operation(x, c, hyperparams)
    return x


def smaller_separable_tran(x, c, k, s, d, name, hyperparams):
    # --- Extreme Reduction
    x = extreme_reduction(x, hyperparams)

    # --- Depthwise Operation
    x = depthwise_upsample_operation(x, k, s, d, hyperparams)
    x = transform.concat(x)

    # --- Pointwise Operation
    x = pointwise_operation(x, c, hyperparams)
    return x


def smallest_separable_conv(x, c, k, s, d, name, hyperparams):
    # --- Extreme Reduction
    x = extreme_reduction(x, hyperparams)

    # --- (Pooling)
    x = learner.conv(x=x, c=1, k=1, s=s, hyperparams=hyperparams)

    # --- Pointwise Operation
    x = pointwise_operation(x, c, hyperparams)
    return x


def smallest_separable_tran(x, c, k, s, d, name, hyperparams):
    # --- Extreme Reduction
    x = extreme_reduction(x, hyperparams)

    # --- Determine if need to pool
    pooling = (
        hyperparams["model"]["pool_type"] != "none"
        and hyperparams["model"]["pool_type"] != "aspp"
        and hyperparams["model"]["pool_type"] != "acsp"
        and hyperparams["model"]["pool_type"] != "wasp"
    )
    if pooling:
        x = learner.tran(x=x, c=1, k=1, hyperparams=hyperparams)
    else:
        x = learner.conv(x=x, c=1, k=1, s=s, hyperparams=hyperparams)

    # --- Pointwise Operation
    x = pointwise_operation(x, c, hyperparams)
    return x



"""Attention-gate block"""

import tfcaidm.models.layers.activation as activation
import tfcaidm.models.layers.learner as learner
import tfcaidm.models.layers.transform as transform
from tfcaidm.models.utils import select as model_select


def attention_gate(x, x_skip, hyperparams, s=(1, 2, 2), **kwargs):
    """Implementation of attention-gate block https://arxiv.org/abs/1804.03999)"""

    c = x.shape[-1]

    # --- Gating signal
    gs = model_select.conv_selection(x=x, c=c, k=1, hyperparams=hyperparams)

    # --- Skip signal
    ss = model_select.conv_selection(x=x_skip, c=c, k=1, s=s, hyperparams=hyperparams)

    # --- Combine signals
    cs = activation.nonlinearity(hyperparams)()(gs + ss)

    # --- Attention
    a = learner.pre_activation_conv(x=cs, c=1, k=1)
    a = activation.attn_msk(a, hyperparams)

    # --- Upsample
    if s != (1, 1, 1):
        a = transform.up_sample(x=a, s=s)

    return x_skip * a



"""Tensorboard viewer for a set of experiments"""

import os
import sys
from pathlib import Path
from argparse import ArgumentParser


def tensorboard_viewer(run_path, port):
    """Function to bring up a tensorboard for a directory of experiments
    Args:
        run_path (str): path to a directory of experiments
        port     (str): port number to run tensorboard
    """

    run_path = Path(run_path).resolve()
    exps = os.listdir(run_path)
    cmd = "tensorboard --logdir_spec="

    for exp in exps:
        if os.path.isdir(run_path / exp):
            cmd += exp + ":" + run_path + "/" + exp + "/" + "logdirs" + ","

    cmd = cmd[:-1]
    cmd += " --bind_all --port " + port
    print("\n- command: {}\n".format(cmd))
    os.system(cmd)


def parser(args):
    p = ArgumentParser(
        description="- tensorboard viewer for all runs under a given experiment"
    )
    p.add_argument(
        "--run_path",
        type=str,
        required=True,
        help="- path to specific experiment ie. run_path=exp/xr_pna/logs/2021-11-03_19-10-43_PDT",
    )
    p.add_argument("--port", type=str, default="9000", help="- port number")
    parsed = p.parse_args(args)
    arguments = {name: getattr(parsed, name) for name in vars(parsed)}

    return arguments


def view(run_path, port, **kwargs):
    tensorboard_viewer(run_path, port)


if __name__ == "__main__":
    args = parser(sys.argv[1:])
    view(**args)



"""Tools to get or measure time"""

import time
import pytz
import datetime
from pytz import timezone

import tfcaidm.common.constants as constants


def get_date():
    date = datetime.datetime.now(tz=pytz.utc)
    date = date.astimezone(timezone("US/Pacific"))
    return date.strftime(constants.date_format)


def timediff(end, start):
    """Measure the time difference between two datetime objects

    Args:
        end (datetime): latest datetime
        start (datetime): earliest datetime

    Returns:
        string: time difference as a string
    """

    end_date = datetime.datetime.strptime(end, constants.date_format)
    start_date = datetime.datetime.strptime(start, constants.date_format)

    return end_date - start_date


def profile(func):
    start = time.time()

    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    end = time.time()
    print(end - start)

    return wrapper



"""Interface for getting available losses"""

from tfcaidm.common.inherit import inherit
from tfcaidm.losses.custom import _zoo


@inherit(_zoo.losses)
def available_losses():
    return custom_loss()


# --- loss in csv
def custom_loss():
    customs = {}
    return customs



"""Entropy related loss functions"""

import tensorflow as tf
from tensorflow import losses
from tensorflow.keras import layers


class SparseCategoricalCrossentropy(layers.Layer):
    """Computes the categorical cross-entropy loss"""

    def __init__(self, name="SparseCategoricalCrossentropy", **kwargs):
        super(SparseCategoricalCrossentropy, self).__init__(name=name, **kwargs)
        self.loss = losses.SparseCategoricalCrossentropy(from_logits=True)

    def call(self, y_true, y_pred, weights=None, add_loss=True, **kwargs):
        loss = self.loss(
            y_true=y_true,
            y_pred=y_pred,
            sample_weight=weights,
        )

        if add_loss:
            self.add_loss(loss)

        return loss


class WeightedCategoricalCrossentropy(layers.Layer):
    """Computes the weighted cross-entropy loss"""

    def __init__(self, name="WeightedCategoricalCrossentropy", **kwargs):
        super(WeightedCategoricalCrossentropy, self).__init__(name=name, **kwargs)
        self.loss = tf.nn.weighted_cross_entropy_with_logits

    def call(
        self, y_true, y_pred, weights=None, class_of_interest=1, add_loss=True, **kwargs
    ):
        # ---- Extract the class of interest
        labels = tf.cast(y_true[..., 0] == class_of_interest, tf.float32)
        logits = tf.cast(y_pred[..., class_of_interest], tf.float32)

        # pos_weight > 1 decreases: false negatives, increases: false positives
        # pos_weight < 1 decreases: false positives, increases: false negatives
        weights = 1.0 if weights is None else weights
        pos_weight = tf.cast(tf.math.reduce_max(weights), tf.float32)

        loss = self.loss(
            labels=labels,
            logits=logits,
            pos_weight=pos_weight,
        )

        if add_loss:
            self.add_loss(loss)

        return loss



"""Nonlinear activation functions"""

import tensorflow as tf
from tensorflow.keras import activations, layers

from tfcaidm.models.layers.regularization import normalization


activation_kwargs = {
    "elu": layers.ELU,
    "relu": layers.ReLU,
    "prelu": layers.PReLU,
    "leaky": layers.LeakyReLU,
    "gelu": lambda: tf.keras.activations.gelu
    # more: https://www.tensorflow.org/api_docs/python/tf/keras/activations
}


def nonlinearity(hyperparams):
    k = hyperparams["model"]["activ"]

    if k not in activation_kwargs:
        raise ValueError(f"ERROR! Activation function `{k}` is not defined!")

    return activation_kwargs[k]


def apply_all_activations(x, hyperparams):
    if hyperparams["model"]["order"] == "rnc":  # relu(norm(conv))
        x = normalization(hyperparams)()(x)
        x = nonlinearity(hyperparams)()(x)
    else:
        x = nonlinearity(hyperparams)()(x)
        x = normalization(hyperparams)()(x)

    return x


# --- Output or attention activations
tanh = lambda x: activations.tanh(x)
sigmoid = lambda x: activations.sigmoid(x)
softmax = lambda x: activations.softmax(x)


def attn_msk(x, hyperparams):
    if hyperparams["model"]["attn_msk"] == "sigmoid":
        return sigmoid(x)

    elif hyperparams["model"]["attn_msk"] == "tanh":
        return (tanh(x) + 1) / 2

    else:
        return softmax(x)



"""Pass and store parameters (dicts) in decorated functions"""

from functools import wraps

err_msg = lambda k: f"Cannot override, `{k}`, choose a different key!"

# --- Decorator with arguments
def inherit(defined_dict):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            custom_dict = {}
            customs = func(*args, **kwargs)

            for k, v in defined_dict.items():
                custom_dict[k] = v

            for k, v in customs.items():
                assert k not in defined_dict, err_msg(k)
                custom_dict[k] = v

            return custom_dict

        return wrapper

    return decorator



"""Distance related metrics"""

from tensorflow.keras import layers
from tensorflow.keras import metrics


class MeanAbsoluteError(layers.Layer):
    """Computes the mean absolute error (L1) metric"""

    def __init__(self, name="MeanAbsoluteError", **kwargs):
        super(MeanAbsoluteError, self).__init__(name=name, **kwargs)
        self.metric = metrics.MeanAbsoluteError()

    def call(self, y_true, y_pred, weights=None, add_metric=True, **kwargs):

        metric = self.metric(y_true=y_true, y_pred=y_pred, sample_weight=weights)

        if add_metric:
            self.add_metric(metric, name=self.metric.name)

        return metric


class MeanSquaredError(layers.Layer):
    """Computes the mean squared error (L2) metric"""

    def __init__(self, name="MeanSquaredError", **kwargs):
        super(MeanSquaredError, self).__init__(name=name, **kwargs)
        self.metric = metrics.MeanSquaredError()

    def call(self, y_true, y_pred, weights=None, add_metric=True, **kwargs):

        metric = self.metric(y_true=y_true, y_pred=y_pred, sample_weight=weights)

        if add_metric:
            self.add_metric(metric, name=self.metric.name)

        return metric



"""Select model blocks to use"""

import tfcaidm.models.custom.registry as registry
import tfcaidm.models.layers.transform as transform
import tfcaidm.common.constants as constants

# --- conv_type in csv
def conv_selection(
    x=None, c=1, k=(1, 1, 1), s=(1, 1, 1), d=(1, 1, 1), name=None, hyperparams={}
):
    """Choose between different types of convolution methods"""

    conv_types = registry.available_convs()

    # --- Extract hyperparams from params csv
    conv_type = hyperparams["model"]["conv_type"]

    if conv_type not in conv_types:
        raise ValueError(f"ERROR! Conv layer `{conv_type}` is not defined!")

    return conv_types[conv_type](
        x=x, c=c, k=k, s=s, d=d, name=name, hyperparams=hyperparams
    )


# --- conv_type in csv (same settings as conv_selection)
def tran_selection(
    x=None, c=1, k=(1, 1, 1), s=(1, 2, 2), d=(1, 1, 1), name=None, hyperparams={}
):
    """Choose between different types of transpose convolution methods"""

    conv_types = registry.available_trans()

    # --- Extract hyperparams from params csv
    conv_type = hyperparams["model"]["conv_type"]

    if conv_type not in conv_types:
        raise ValueError(f"ERROR! Conv layer `{conv_type}` is not defined!")

    return conv_types[conv_type](
        x=x, c=c, k=k, s=s, d=d, name=name, hyperparams=hyperparams
    )


# ---pool_type in csv
def pool_selection(x, c, k, s, hyperparams):
    """Choose between different pooling layers"""

    pool_types = registry.available_pools()

    # --- Extract hyperparams from params csv
    pool_type = hyperparams["model"]["pool_type"]

    if pool_type not in pool_types:
        raise ValueError(f"ERROR! Pool layer `{pool_type}` is not defined!")

    return pool_types[pool_type](x=x, c=c, k=k, s=s, hyperparams=hyperparams)


# --- eblock in csv
def encoder_selection(x, c, k, s, hyperparams):
    """Choose between different encoder feature extraction blocks"""

    eblocks = registry.available_encoders()

    # --- Extract hyperparams from params csv
    eblock = hyperparams["model"]["eblock"].split("_")[0]

    if eblock not in eblocks:
        raise ValueError(f"ERROR! Encoder block `{eblock}` is not defined!")

    return eblocks[eblock](x=x, c=c, k=k, s=s, hyperparams=hyperparams)


# --- dblock in csv
def decoder_selection(x, x_skip, c, k, s, hyperparams):
    """Choose between different decoder feature extraction blocks"""

    dblocks = registry.available_decoders()

    # --- Extract hyperparams from params csv
    dblock = hyperparams["model"]["dblock"]

    if dblock not in dblocks:
        raise ValueError(f"ERROR! Decoder block `{dblock}` is not defined!")

    return dblocks[dblock](
        x=x,
        x_skip=x_skip,
        c=c,
        k=k,
        s=s,
        hyperparams=hyperparams,
    )


def head_selection(x, hyperparams):
    """Choose between different output heads"""

    heads = registry.available_heads()

    # --- Extract hyperparams from params csv
    head = hyperparams["head"]

    if head not in heads:
        raise ValueError(f"ERROR! Model head `{head}` is not defined!")

    return heads[head](**x)


def task_selection(x, inputs, client, task="auto"):
    """Choose between different task(s)"""

    task_types = registry.available_tasks()

    if task not in task_types:
        raise ValueError(f"ERROR! Model tasks `{task}` is not defined!")

    return task_types[task](x, inputs, client)


# --- model in csv
def model_selection(x, hyperparams):
    """Choose between different model architectures"""

    model_archs = registry.available_models()

    # --- Extract hyperparams from params csv
    model_arch = hyperparams["model"]["model"]
    n = hyperparams["model"]["depth"]
    c = hyperparams["model"]["width"]
    k = hyperparams["model"]["kernel_size"]
    s = hyperparams["model"]["strides"]

    if model_arch not in model_archs:
        raise ValueError(f"ERROR! Model `{model_arch}` is not defined!")

    return model_archs[model_arch](x, n, c, k, s, hyperparams)


def input_selection(x, name, hyperparams):
    """Concatenate and project additional input features"""

    # --- Choose first input when not provided
    if name is None:
        names = [*hyperparams["train"]["xs"].keys()]
        assert len(names), "ERROR! Issue building inputs, check your client yaml file!"
        name = names[0]

    # --- Extract hyperparams from params csv
    addons = hyperparams["train"]["xs"][name]
    c = hyperparams["model"]["width"]

    features = [name]

    if type(addons) == dict:
        for k in addons:
            if type(addons[k]) == dict:
                features += [addons[k][n] for n in addons[k] if n == "name"]

    x_new = [x[i] for i in x if i in features]
    x_new = transform.concat(x_new)

    return conv_selection(x=x_new, c=c, k=1, s=1, hyperparams=hyperparams)



"""Model head interface"""

from tfcaidm.models.layers import learner
from tfcaidm.models.layers import transform


class Decoder:
    """Determines the decoder head to use for a given output"""

    @classmethod
    def last_layer(cls, decoders, *args, **kwargs):
        return decoders[-1]

    @classmethod
    def multi_scale(cls, decoders, *args, **kwargs):
        # Assumes 5D tensor of [batch, depth, width, height, channels]
        layers = []

        final_shape = decoders[-1].shape
        f_d, f_w, f_h, f_c = final_shape[1:]

        for decoder in decoders:
            inter_shape = decoder.shape
            i_d, i_w, i_h = inter_shape[1:-1]

            s = (int(f_d / i_d), int(f_w / i_w), int(f_h / i_h))

            decoder = learner.pre_activation_conv(x=decoder, c=f_c, k=1)

            layer = transform.up_sample(decoder, s=s)
            layers.append(layer)

        layer = transform.add(layers)

        return layer

    @classmethod
    def deep_supervision(cls, full_res, *args, **kwargs):
        layer = transform.add(full_res)
        return layer

    @classmethod
    def complex_supervision(cls, full_res, decoders, *args, **kwargs):
        decoders = cls.multi_scale(decoders)
        full_res = cls.deep_supervision(full_res)
        layer = decoders + full_res
        return layer


class Encoder:
    """Determines the encoder head to use for a given output"""

    @classmethod
    def last_layer(cls, encoders, *args, **kwargs):
        return encoders[-1]

    @classmethod
    def multi_scale(cls, encoders, *args, **kwargs):
        # Assumes 5D tensor of [batch, depth, width, height, channels]
        layers = []

        final_shape = encoders[-1].shape
        f_d, f_w, f_h = final_shape[1:-1]
        f_c = encoders[0].shape[-1]

        for encoder in encoders:
            inter_shape = encoder.shape
            i_d, i_w, i_h = inter_shape[1:-1]

            s = (int(i_d / f_d), int(i_w / f_w), int(i_h / f_h))

            encoder = learner.pre_activation_conv(x=encoder, c=f_c, k=1)

            layer = transform.average_pool(encoder, s=s) * 0.5
            layer += transform.max_pool(encoder, s=s) * 0.5
            layers.append(layer)

        layer = transform.add(layers)

        return layer



"""Training hyperparameter interface"""

import numpy as np
from pathlib import Path


import tfcaidm.common.timedate as timedate
from tfcaidm.models.model import Model
from tfcaidm.data.dataset import Dataset
from tfcaidm.jobs.utils.params import HyperParameters
from tfcaidm.train.utils.select import callback_selection
from tfcaidm.train.utils.summary import Summary, save_results


class Trainer(HyperParameters):
    def __init__(self, hyperparams):
        HyperParameters.__init__(self, hyperparams)
        self.timestamp = timedate.get_date()

    def cross_validation(self, n_folds=None, save=False, callbacks=[]):
        """K-fold cross validation (up to K=5)

        Args:
            n_folds (int): Number of cross validation folds. Defaults to hyperparams["train"]["trainer"]["n_folds"]
            save (bool): Save client and model after model training. Defaults to False
            callbacks (List, None):
                - if None, disable callbacks
                - if non-empty, use callbacks from list
                - if emtpy list, use callbacks from hyperparams["train"]["trainer"]["callbacks"]

        Returns:
            dict: Returns a dictionary containing outputs from model.fit and model.eval
        """

        n_folds = choose(n_folds, self.results["train"]["trainer"]["n_folds"])
        assert n_folds > 0, "ERROR! n_folds must be greater than 0!"

        histories = {
            "train_history": [],
            "train_results": [],
            "valid_results": [],
        }

        # --- Cross validation generator
        cv = self.__cross_valid(n_folds, save, callbacks)

        for history in cv:
            hist = history
            train_history, train_results, valid_results = hist

            # --- Accumulate results
            histories["train_history"].append(train_history.history)
            histories["train_results"].append(train_results)
            histories["valid_results"].append(valid_results)

        return histories

    def __cross_valid(self, n_folds, save, callbacks):
        # --- Cross validation loop
        for fold in range(n_folds):

            client = Dataset(self.hyperparams).get_client(fold)

            # --- Load train dataset
            gen_train, gen_valid = client.create_generators(test=False)

            # --- Load validation dataset
            gen_train_test, gen_valid_test = client.create_generators(test=True)

            # --- Create a model
            model = Model(client).create()

            # --- Train and evaluate model
            train_history = self.fit(model, gen_train, gen_valid, callbacks=callbacks)
            train_results = self.eval(model, gen_train_test)
            valid_results = self.eval(model, gen_valid_test)

            if save:
                self.__checkpoint(client, model, fold)

            history = [train_history, train_results, valid_results]

            yield history

    def __checkpoint(self, client, model, fold=0):
        output_dir = self.hyperparams["train"]["trainer"]["log_dir"]
        client_name = Path(self.hyperparams["env"]["path"]["client"]).stem
        model_name = self.hyperparams["model"]["model"]

        client_path = output_dir + "/" + client_name + f"_{fold}" + ".yml"
        model_path = output_dir + "/" + model_name + f"_{fold}"

        self.save_client(client, client_path)
        self.save_model(model, model_path)

    def eval(self, model, gen_data, verbose=0):
        results = model.evaluate(
            x=gen_data,
            batch_size=1,
            verbose=verbose,
            return_dict=True,
        )

        return results

    def fit(
        self,
        model,
        gen_train,
        gen_valid,
        *,
        iters=None,
        steps_per_epoch=None,
        validation_freq=None,
        callbacks=[],
    ):
        """Model training

        Args:
            model (TFModel): A compiled tensorflow model
            gen_train (generator): Training dataset generator
            gen_valid (generator): Validation dataset generator
            iters (int): Total number of training iterations. Defaults to hyperparams["train"]["trainer"]["iters"]
            steps_per_epoch (int): Number of forward passes per epoch. Defaults to hyperparams["train"]["trainer"]["steps"]
            validation_freq (int): Run model validation every N epochs. Defaults to hyperparams["train"]["trainer"]["valid_freq"]
            callbacks (List, None):
                - if None, disable callbacks
                - if non-empty, use callbacks from list
                - if emtpy list, use callbacks from hyperparams["train"]["trainer"]["callbacks"]

        Returns:
            model.history: A model.history object
        """

        assert (iters is None and steps_per_epoch is None) or (
            iters is not None and steps_per_epoch is not None
        ), "ERROR! Must define both or neither `iters` and `steps_per_epoch`..."

        iters = choose(iters, self.results["train"]["trainer"]["iters"])
        steps_per_epoch = choose(
            steps_per_epoch, self.results["train"]["trainer"]["steps"]
        )
        validation_freq = choose(
            validation_freq, self.results["train"]["trainer"]["valid_freq"]
        )

        epochs = int(iters / steps_per_epoch)

        if callbacks is not None and len(callbacks) == 0:
            funcs = callback_selection(self.hyperparams)
            callbacks = [func(self.hyperparams) for func in funcs]

        history = model.fit(
            x=gen_train,
            epochs=epochs,
            steps_per_epoch=steps_per_epoch,
            validation_data=gen_valid,
            validation_steps=steps_per_epoch,
            validation_freq=validation_freq,
            callbacks=callbacks,
        )

        # --- Save number of model params
        self.results["model"]["num_params"] = Model.get_num_params(model)

        return history

    def save_client(self, client, path):
        client.to_yml(path)

    def save_model(self, model, path):
        model.save(path)

    def save_inference(self, model, gen_train, gen_valid, output_dir, fold):
        train_output = output_dir + "/" + "train" + f"_{fold}" + ".npz"
        valid_output = output_dir + "/" + "valid" + f"_{fold}" + ".npz"

        self.save_outputs(model, gen_train, path=train_output)
        self.save_outputs(model, gen_valid, path=valid_output)

    def save_outputs(self, model, gen_data, path):
        array = lambda list_: np.array(list_ if len(list_) else [0])

        data = {
            "xs": [],
            "ys": [],
            "zs": [],
        }

        for xs, ys in gen_data:
            zs = model.predict(xs)
            data["xs"].append(xs)
            data["ys"].append(ys)
            data["zs"].append(zs)

        data = {k: array(v) for k, v in data.items()}
        np.savez_compressed(path, **data)

    def save_results(self, histories, name="val_loss"):
        assert (
            type(histories) == dict
        ), "ERROR! Argument `histories` must be of type dict!"

        valid_freq = self.results["train"]["trainer"]["valid_freq"]

        summary = Summary(histories["train_history"])
        results = {"history": summary.summarize_best_history(valid_freq, name)}
        self.update_results(results)

        summary = Summary(histories["train_results"])
        results = {"train_eval": summary.summarize_eval()}
        self.update_results(results)

        summary = Summary(histories["valid_results"])
        results = {"valid_eval": summary.summarize_eval()}
        self.update_results(results)

        start_time = self.timestamp
        end_time = timedate.get_date()

        self.results["train"]["trainer"]["train_time"] = timedate.timediff(
            end_time, start_time
        )

        save_results(self.flatten(self.results))

    def update_results(self, new_results):
        new_results = self.flatten(new_results)
        results = self.flatten(self.results)
        results.update(new_results)

        results = self.unflatten(results)
        self.results.update(results)

    @staticmethod
    def load_outputs(path):
        return np.load(path, allow_pickle=True)


def choose(a, b):
    if a is None:
        a = b
    else:
        b = a
    return a



"""Adds positional information to features"""

import numpy as np


def get_coords(arrays, shift=0, epsilon=1e-9):
    """Get a normalized coordinate map, default [-1, 1]

    For a tensor with dim:
        - (B, D, W, H, C) => CoordMap(D, W, H)
        - (B, W, H, C) => CoordMap(W, H)
        - (B, H, C) => CoordMap(H)

    Args:
        arrays (np.array): A numpy array of dim > 1
        shift (int): An optional value to shift coordinate system
        epsilon (float): Added for numerical stability

    Returns:
        np.array: Normalized coordinate map(s)
    """

    assert (
        len(arrays.shape) > 2
    ), "ERROR! np.array.shape > 2, must be at least (B, H, C)"

    # --- Get input dims
    dim_c = 1
    dim_b = len(arrays.shape) - 1
    shape = arrays.shape[-dim_b:-dim_c]

    dim_1 = dim_c
    dim_n = len(shape)

    # --- Helpers to define the coordinate transform
    coord = lambda x: np.arange(x)
    mid = lambda x: (x - 1) / 2
    norm = lambda x, x_mid: (x - x_mid) / (x_mid + epsilon)  # [-1, 1]
    ncoord = lambda x: norm(coord(x), mid(x))

    mesh = [ncoord(shape[-i]) for i in range(dim_1, dim_n + 1)]

    # --- Get coordinate maps
    maps = np.array(np.meshgrid(*mesh)) + shift

    # --- Rearrange coordinate maps
    axes = rearrange(maps.shape)
    coord_maps = np.transpose(maps, axes=axes)

    return coord_maps


def rearrange(shape):
    """Get's axis needed to rearrange np.array

    Args:
        shape (tuple): np.array.shape

    Returns:
        tuple: axis in which to rearrange np.array
    """

    n_dim = len(shape)
    dims = []

    for i in range(n_dim - 1, 0, -1):

        if n_dim - len(dims) == 2:
            dims.append(1)
            dims.append(0)
            break

        elif n_dim - len(dims) == 3:
            dims.append(1)
            dims.append(2)
            dims.append(0)
            break

        else:
            dims.append(i)

    return dims


# --- Return coordinate maps
def add_coordinates(arrays, hyperparams, **kwargs):
    coord_maps = get_coords(arrays)
    return coord_maps



"""Model encoder, decoder, and pooling blocks"""

from tfcaidm.models.utils import select as model_select


def in_conv(x, c, k, hyperparams):
    x = model_select.conv_selection(x=x, c=c, k=k, hyperparams=hyperparams)

    return x


def handle_pooling(s, d=None):
    """Handle pooling size to ensure valid feature map size"""

    if d is not None:
        size = lambda s, d: 1 if d < 0 else s
        s = [size(s_i, d_i) for s_i, d_i in zip(s, d)]

    return s


def residual(x, c, k, s, hyperparams):
    """Implementation of resnet block"""

    x = model_select.conv_selection(x=x, c=c, k=k, s=s, hyperparams=hyperparams)

    for _ in range(hyperparams["model"]["elayer"] - 1):
        x += model_select.encoder_selection(x=x, c=c, k=k, s=s, hyperparams=hyperparams)

    return x


def down_conv(x, c, k, s, hyperparams):
    """Encoder module"""

    c = round(c * hyperparams["model"]["width_scaling"])
    s = handle_pooling(s)

    x = residual(x=x, c=c, k=k, s=1, hyperparams=hyperparams)
    x = model_select.pool_selection(x=x, c=c, k=k, s=s, hyperparams=hyperparams)

    return x, c


def up_conv(x, x_skip, c, k, s, d, hyperparams):
    """Decoder module"""

    c = round(c / hyperparams["model"]["width_scaling"])
    s = handle_pooling(s, d)

    x_skip = model_select.decoder_selection(
        x=x, x_skip=x_skip, c=c, k=k, s=s, hyperparams=hyperparams
    )
    x = model_select.tran_selection(x=x, c=c, k=k, s=s, hyperparams=hyperparams)
    x = model_select.conv_selection(
        x=(x + x_skip), c=c, k=k, s=1, hyperparams=hyperparams
    )

    return x, c



"""Dice score related loss functions"""

import tensorflow as tf
from tensorflow.keras import layers

from tfcaidm.losses.funcs.entropy import SparseCategoricalCrossentropy


reduce = lambda x: tf.math.reduce_sum(x)


class DiceLoss(layers.Layer):
    """Computes the soft dice-score (F1) loss"""

    def __init__(self, name="DiceLoss", **kwargs):
        super(DiceLoss, self).__init__(name=name, **kwargs)

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        class_of_interest=1,
        epsilon=1e-6,
        add_loss=True,
        **kwargs,
    ):
        # ---- Extract the class of interest
        true = tf.cast(y_true[..., 0] == class_of_interest, tf.float32)
        pred = tf.nn.softmax(y_pred, axis=-1)[..., class_of_interest]

        if weights is not None:
            true *= tf.cast(weights[..., 0], tf.float32)
            pred *= tf.cast(weights[..., 0], tf.float32)

        A = reduce(true * pred) * 2
        B = reduce(true) + reduce(pred) + epsilon

        C = reduce(true * true)
        D = reduce(true) + epsilon

        loss = (C / D) - (A / B)

        if add_loss:
            self.add_loss(loss)

        return loss


class LogCoshDiceLoss(DiceLoss):
    """Computes the soft logcosh dice-score (F1) loss"""

    def __init__(self, name="LogCoshDiceLoss", **kwargs):
        super(LogCoshDiceLoss, self).__init__(name=name, **kwargs)

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        class_of_interest=1,
        epsilon=1e-6,
        add_loss=True,
        **kwargs,
    ):
        loss = DiceLoss.call(
            self,
            y_true=y_true,
            y_pred=y_pred,
            weights=weights,
            class_of_interest=class_of_interest,
            epsilon=epsilon,
            add_loss=False,
        )
        cosh = lambda x: (tf.exp(x) + tf.exp(-x)) / 2.0

        loss = tf.math.log(cosh(loss))

        if add_loss:
            self.add_loss(loss)

        return loss


class DiceCrossentropyLoss(SparseCategoricalCrossentropy, DiceLoss):
    """Computes a weighted dice-score cross-entropy loss"""

    def __init__(self, name="DiceCrossentropyLoss", **kwargs):
        super(DiceCrossentropyLoss, self).__init__(name=name, **kwargs)

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        class_of_interest=1,
        epsilon=1e-6,
        alpha=0.5,
        add_loss=True,
        **kwargs,
    ):
        entropy = SparseCategoricalCrossentropy.call(
            y_true=y_true,
            y_pred=y_pred,
            sample_weight=weights,
            add_loss=False,
        )

        dice = DiceLoss.call(
            self,
            y_true=y_true,
            y_pred=y_pred,
            weights=weights,
            class_of_interest=class_of_interest,
            epsilon=epsilon,
            add_loss=False,
        )

        loss = alpha * dice + (1 - alpha) * entropy

        if add_loss:
            self.add_loss(loss)

        return loss


class MultiDiceLoss(DiceLoss):
    """Computes the soft dice-score (F1) loss over several classes"""

    def __init__(self, name="MultiDiceLoss", **kwargs):
        super(MultiDiceLoss, self).__init__(name=name, **kwargs)

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        classes_of_interest=[0, 1],
        epsilon=1e-6,
        add_loss=True,
        **kwargs,
    ):

        losses = []
        for class_of_interest in classes_of_interest:
            loss = DiceLoss.call(
                self,
                y_true=y_true,
                y_pred=y_pred,
                weights=(weights, None)[class_of_interest == 0],
                class_of_interest=class_of_interest,
                epsilon=epsilon,
                add_loss=False,
            )

            losses.append(loss)

        loss = sum([loss for loss in losses]) / len(losses)

        if add_loss:
            self.add_loss(loss)

        return loss



"""Dataset hyperparameter interface"""

from tfcaidm.data.jclient import JClient
from tfcaidm.jobs.utils.params import HyperParameters


class Dataset(HyperParameters):
    def __init__(self, hyperparams):
        super(Dataset, self).__init__(hyperparams=hyperparams)

    def get_client(self, fold):
        path = self.hyperparams["env"]["path"]["client"]

        configs = {
            "batch": {
                "size": self.hyperparams["train"]["trainer"]["batch_size"],
                "fold": fold,
            },
        }

        client = JClient(path, hyperparams=self.hyperparams, configs=configs)

        return client



"""Basic neural network layers"""

from tensorflow.keras import layers

from tfcaidm.models.layers.activation import apply_all_activations


def conv_layer(c, k=(1, 3, 3), s=(1, 1, 1), d=1, name=None):
    layer = layers.Conv3D(
        filters=c, kernel_size=k, strides=s, dilation_rate=d, name=name, padding="same"
    )

    return layer


def tran_layer(c, k=(1, 3, 3), s=(1, 2, 2), d=1, name=None):
    layer = layers.Conv3DTranspose(
        filters=c, kernel_size=k, strides=s, dilation_rate=d, name=name, padding="same"
    )

    return layer


def pre_activation_conv(x, c, k=(1, 3, 3), s=(1, 1, 1), d=1, name=None):
    x = conv_layer(c=c, k=k, s=s, d=d, name=name)(x)

    return x


def pre_activation_tran(x, c, k=(1, 3, 3), s=(1, 2, 2), d=1, name=None):
    x = tran_layer(c=c, k=k, s=s, d=d, name=name)(x)

    return x


def pre_activation_mlp(x, c, name=None):
    x = layers.Dense(units=c, name=name)(x)

    return x


def conv(x, c, hyperparams, k=(1, 3, 3), s=(1, 1, 1), d=1, name=None, **kwargs):
    x = pre_activation_conv(x, c, k=k, s=s, d=d, name=name)
    x = apply_all_activations(x, hyperparams)

    return x


def tran(x, c, hyperparams, k=(1, 3, 3), s=(1, 2, 2), d=1, name=None, **kwargs):
    x = pre_activation_tran(x, c, k=k, s=s, d=d, name=name)
    x = apply_all_activations(x, hyperparams)

    return x


def mlp(x, c, hyperparams, name=None, **kwargs):
    x = pre_activation_mlp(x, c, name=name)
    x = apply_all_activations(x, hyperparams)

    return x



"""Interface for getting available model blocks"""

from tfcaidm.common.inherit import inherit
from tfcaidm.models.custom import _zoo


@inherit(_zoo.conv_types)
def available_convs():
    return custom_conv()


@inherit(_zoo.tran_types)
def available_trans():
    return custom_tran()


@inherit(_zoo.pool_types)
def available_pools():
    return custom_pool()


@inherit(_zoo.eblocks)
def available_encoders():
    return custom_encoder()


@inherit(_zoo.dblocks)
def available_decoders():
    return custom_decoder()


@inherit(_zoo.heads)
def available_heads():
    return custom_head()


@inherit(_zoo.task_types)
def available_tasks():
    return custom_task()


@inherit(_zoo.models)
def available_models():
    return custom_model()


# ---- Functions to overload

# --- conv_type in csv
def custom_conv():
    customs = {}
    return customs


# --- conv_type in csv
def custom_tran():
    customs = {}
    return customs


# ---pool_type in csv
def custom_pool():
    customs = {}
    return customs


# --- eblock in csv
def custom_encoder():
    customs = {}
    return customs


# --- dblock in csv
def custom_decoder():
    customs = {}
    return customs


# --- head in csv
def custom_head():
    customs = {}
    return customs


def custom_task():
    customs = {}
    return customs


# --- model in csv
def custom_model():
    customs = {}
    return customs


# --- example model hyperparams
def sample_params():
    return {
        "model": {
            "model": "unet",
            "conv_type": "conv",
            "pool_type": "conv",
            "eblock": "conv",
            "elayer": 1,
            "dblock": "conv",
            "depth": 4,
            "width": 32,
            "width_scaling": 1,
            "kernel_size": [1, 3, 3],
            "strides": [1, 2, 2],
            "bneck": 2,
            "branches": 4,
            "atrous_rate": 6,
            "order": "rnc",
            "norm": "bnorm",
            "activ": "leaky",
            "attn_msk": "softmax",
        }
    }



"""UNet model"""

import math

import tfcaidm.models.layers.extra as extra
import tfcaidm.models.nn.utils as utils


def max_downsamples(x_dim, s_dim, n):
    """Determine amount of possible downsamples"""

    try:
        return int(math.log(x_dim, s_dim)) - n
    except:
        return 1


def unet(x, n, c, k, s, hyperparams):
    """Implementation of unet model"""

    # --- Control downsampling
    x_size = x.shape[1:-1]  # depth, width, height
    d_size = [
        max_downsamples(x_size[0], s[0], n),
        max_downsamples(x_size[1], s[1], n),
        max_downsamples(x_size[2], s[2], n),
    ]

    # --- Refine input
    x = utils.in_conv(x, c, k, hyperparams)
    x = extra.layer_name(x, "eblock_0")
    print(x.shape)

    # --- Store intermediate layers
    e_list = [x]
    d_list = []

    # --- Encoder network
    for i in range(n):
        x, c = utils.down_conv(x, c, k, s, hyperparams)
        x = extra.layer_name(x, f"eblock_{i + 1}")

        e_list.append(x)
        print(x.shape)

    # --- Decoder network
    for i in range(n):
        x_skip = e_list[-(i + 2)]
        x, c = utils.up_conv(x, x_skip, c, k, s, d_size, hyperparams)
        x = extra.layer_name(x, f"dblock_{i}")

        d_list.append(x)
        d_size = [d_i + 1 if d_i else d_i for d_i in d_size]
        print(x.shape)

    return {
        "encoders": e_list,
        "decoders": d_list,
        "full_res": [e_list[0], d_list[-1]],
    }



"""U^2Net block"""

import math

import tfcaidm.models.layers.activation as activation
import tfcaidm.models.layers.learner as learner
import tfcaidm.models.layers.transform as transform
import tfcaidm.models.layers.regularization as regularization
from tfcaidm.models.utils import select as model_select
from tfcaidm.models.nn import utils as nn_utils


def down_conv(x, conv, norm, acti, s, hyperparams):
    """Encoder module"""

    x = layer(x, conv, norm, acti, hyperparams)
    x = transform.average_pool(x=x, s=s)
    return x


def up_conv(x, x_skip, conv, norm, acti, s, d, hyperparams):
    """Decoder module"""

    x = layer(x, conv, norm, acti, hyperparams)
    s = nn_utils.handle_pooling(s, d)
    x = transform.up_sample(x=x, s=s)
    x = layer(x + x_skip, conv, norm, acti, hyperparams)
    return x


def layer(x, conv, norm, acti, hyperparams):
    """Forward pass conv"""

    if hyperparams["model"]["order"] == "rnc":
        x = acti(norm(conv(x)))
    else:
        x = norm(acti(conv(x)))

    return x


def max_downsamples(x_dim, s_dim, n):
    """Determine amount of possible downsamples"""

    try:
        return int(math.log(x_dim, s_dim)) - n
    except:
        return 1


def u2net(x, c, k, s, hyperparams, **kwargs):
    """Implementation of u2net block to achieve the behavior of U2-Net https://arxiv.org/abs/2005.09007"""

    # --- Control downsampling
    n = hyperparams["model"]["branches"]
    x_size = x.shape[1:-1]  # depth, width, height
    d_size = [
        max_downsamples(x_size[0], s[0], n),
        max_downsamples(x_size[1], s[1], n),
        max_downsamples(x_size[2], s[2], n),
    ]

    # --- Shared conv layers
    conv = learner.conv_layer(c, k)
    norm = regularization.normalization(hyperparams)()
    acti = activation.nonlinearity(hyperparams)()

    # --- Store intermediate layers
    x = model_select.conv_selection(x=x, c=c, k=1, hyperparams=hyperparams)
    x_list = [x]

    # --- Encoder network
    for i in range(n):
        x = down_conv(x, conv, norm, acti, s, hyperparams)
        x_list.append(x)

    # --- Decoder network
    for i in range(n):
        x_skip = x_list[-(i + 2)]
        x = up_conv(x, x_skip, conv, norm, acti, s, d_size, hyperparams)
        d_size = [d_i + 1 if d_i else d_i for d_i in d_size]

    del x_list

    return model_select.conv_selection(x=x, c=c, k=k, hyperparams=hyperparams)



"""A crazy inception block"""

# NOTE: Send a PR if you can shorten imports!

import tfcaidm.models.layers.transform as transform
from tfcaidm.models.utils import select as model_select

import tfcaidm.models.blocks.atrous_block as atrous_block
import tfcaidm.models.blocks.atrous_block as atrous_block
import tfcaidm.models.blocks.cbam_block as cbam_block
import tfcaidm.models.blocks.csp_block as csp_block
import tfcaidm.models.blocks.dense_block as dense_block
import tfcaidm.models.blocks.eca_block as eca_block
import tfcaidm.models.blocks.psp_block as psp_block
import tfcaidm.models.blocks.se_block as se_block
import tfcaidm.models.blocks.u2net_block as u2net_block


def inception(x, c, k, hyperparams, **kwargs):
    """Modification of the inception block with dimensionality reduction https://arxiv.org/abs/1409.4842"""

    x_list = [x]
    x_shared = model_select.conv_selection(x=x, c=1, k=1, hyperparams=hyperparams)

    blocks = {
        "aspp": atrous_block.aspp,
        "cbam": cbam_block.cbam,
        "csp": csp_block.csp,
        "dense": dense_block.dense,
        "eca": eca_block.eca,
        "psp": psp_block.psp,
        "se": se_block.se,
        "u2net": u2net_block.u2net,
    }

    # --- Inception network over models/blocks/*
    for k in blocks:
        x_list.append(blocks[k](x_shared, c, k, hyperparams))
    x_list.append(
        model_select.conv_selection(x=x_shared, c=c, k=k, hyperparams=hyperparams)
    )

    # --- Combine feature maps
    x = transform.add(x_list)

    # --- Refine features
    x = model_select.conv_selection(x=x, c=c, k=1, hyperparams=hyperparams)

    return x



__version__ = "0.0.0a2"

from tfcaidm.data.dataset import Dataset
from tfcaidm.data.jclient import JClient
from tfcaidm.jobs.job import Jobs
from tfcaidm.losses.loss import Loss
from tfcaidm.metrics.metric import Metric
from tfcaidm.models.model import Model
from tfcaidm.train.trainer import Trainer



# https://python-packaging.readthedocs.io/en/latest/minimal.html
# https://packaging.python.org/guides/distributing-packages-using-setuptools/

from setuptools import setup, find_packages


with open("README.md", "r") as f:
    long_description = f.read()

packages = find_packages(include=["tfcaidm", "tfcaidm.*"])
install_requires = [
    lib.strip()
    for lib in open("/Users/owner/Code/ai/ml/tfcaidm-pkg/requirements.txt").readlines()
]

setup(
    name="tfcaidm",
    version="0.0.0a2",
    license="GPLv3",
    author="Brandhsu",
    author_email="brandondhsu@gmail.com",
    url="https://github.com/Brandhsu/tfcaidm",
    description="Deep learning pipeline for medical imaging",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=[
        "deep learning",
        "neural networks",
        "medical imaging",
        "computer vision",
    ],
    packages=packages,
    python_requires=">=3.6, <3.8",
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.7",
    ],
)



"""Convolutional block attention (spatial + channel attention)"""

from tensorflow.keras import layers

import tensorflow as tf
import tfcaidm.models.blocks.eca_block as eca_block
import tfcaidm.models.blocks.se_block as se_block
import tfcaidm.models.layers.activation as activation
import tfcaidm.models.layers.learner as learner
import tfcaidm.models.layers.transform as transform


def channel_attention(x, c, r, hyperparams, **kwargs):
    """Implementation of channel-attention module"""

    attention = {
        "cbam_se": se_block.se,
        "cbam_eca": eca_block.eca,
    }

    eblock = hyperparams["model"]["eblock"]

    if eblock not in attention:
        raise ValueError(f"ERROR! Model block `{eblock}` is not defined!")

    return attention[eblock](x, c, hyperparams, r=r)


def spatial_attention(x, k, hyperparams, **kwargs):
    """Implementation of spatial-attention module"""

    # --- Squeeze (channel pool)
    favg = tf.reduce_mean(x, keepdims=True, axis=-1)
    fmax = tf.reduce_max(x, keepdims=True, axis=-1)

    # --- 7x7 receptive field
    x = transform.concat([favg, fmax])

    for _ in range(3):
        x = learner.pre_activation_conv(x=x, c=1, k=k)

    return activation.attn_msk(x, hyperparams)


def cbam(x, c, k, hyperparams, r=4, **kwargs):
    """Implementation of convolutional block attention https://arxiv.org/abs/1807.06521"""

    ca = channel_attention(x, c, r, hyperparams)
    cs = spatial_attention(ca, k, hyperparams)
    return x + ca * cs



"""UNet3+ model"""

import math

import tfcaidm.models.layers.extra as extra
import tfcaidm.models.layers.transform as transform
import tfcaidm.models.nn.utils as utils
from tfcaidm.models.utils import select as model_select


def skip_connections(x_list, c, k, hyperparams):
    """Adds all intermediate skip connections"""

    c = round(c / hyperparams["model"]["width_scaling"])

    x_skip = []
    x_size = x_list[-1].shape[1:-1]

    for i in range(len(x_list)):
        x = x_list[i]
        y_size = x.shape[1:-1]

        # --- Calculate required strides from different scales
        s = [int(math.ceil(y / x)) for x, y in zip(x_size, y_size)]

        # --- Apply pooling over intermediate skip connections
        x = model_select.pool_selection(x=x, c=c, k=k, s=s, hyperparams=hyperparams)
        x_skip.append(x)

    # --- NOTE: Can aggregate using attention, add, or concat
    x_skip = transform.add(x_skip)
    x_skip = model_select.conv_selection(x=x_skip, c=c, k=k, hyperparams=hyperparams)

    return x_skip


def max_downsamples(x_dim, s_dim, n):
    """Determine amount of possible downsamples"""

    try:
        return int(math.log(x_dim, s_dim)) - n
    except:
        return 1


def unet3p(x, n, c, k, s, hyperparams):
    """Implementation of unet 3+ (unet+++) model"""

    # --- Control downsampling
    x_size = x.shape[1:-1]  # depth, width, height
    d_size = [
        max_downsamples(x_size[0], s[0], n),
        max_downsamples(x_size[1], s[1], n),
        max_downsamples(x_size[2], s[2], n),
    ]

    # --- Refine input
    x = utils.in_conv(x, c, k, hyperparams)
    x = extra.layer_name(x, "eblock_0")
    print(x.shape)

    # --- Store intermediate layers
    e_list = [x]
    d_list = []

    # --- Encoder network
    for i in range(n):
        x, c = utils.down_conv(x, c, k, s, hyperparams)
        x = extra.layer_name(x, f"eblock_{i + 1}")

        e_list.append(x)
        print(x.shape)

    # --- Decoder network
    for j in range(n):
        x_skip = skip_connections(e_list[: n - j], c, k, hyperparams)
        x, c = utils.up_conv(x, x_skip, c, k, s, d_size, hyperparams)
        x = extra.layer_name(x, f"dblock_{j}")

        d_list.append(x)
        d_size = [d_i + 1 if d_i else d_i for d_i in d_size]
        print(x.shape)

    return {
        "encoders": e_list,
        "decoders": d_list,
        "full_res": [e_list[0], d_list[-1]],
    }



"""Defined metrics"""

import tfcaidm.metrics.funcs.acc as acc
import tfcaidm.metrics.funcs.dice as dice
import tfcaidm.metrics.funcs.distance as distance

metrics = {
    "acc": acc.Accuracy,
    "bacc": acc.BalancedAccuracy,
    "mae": distance.MeanAbsoluteError,
    "mse": distance.MeanSquaredError,
    "dice": dice.DiceScore,
}



"""Model hyperparameter interface"""

from tensorflow import optimizers
from tensorflow.keras import Input, Model as TFModel, models as TFmodels

from tfcaidm.models.utils import select as model_select
from tfcaidm.common.reproducibility import set_determinism


class Model:
    def __init__(self, client):
        self.client = client
        self.data = client.get_inputs(Input)
        self.hyperparams = client.hyperparams

        self.loss = None
        self.metrics = None
        self.optimizer = None

        set_determinism(seed=self.hyperparams["train"]["trainer"]["seed"])

    def inputs(self, input_name=None):
        return model_select.input_selection(self.data, input_name, self.hyperparams)

    def backbone(self, features):
        return model_select.model_selection(features, self.hyperparams)

    def outputs(self, features):
        return model_select.task_selection(features, self.data, self.client)

    def assemble(self, inputs, outputs):
        return TFModel(inputs=inputs, outputs=outputs)

    def build(self, input_name=None):
        """Build entire model"""

        # --- Select model inputs
        x = self.inputs(input_name)

        # --- Select a model
        x = self.backbone(x)

        # --- Select the model task
        y = self.outputs(x)

        return self.assemble(inputs=self.data, outputs=y)

    def compile(self, model):
        """Compile model with objective functions"""

        lr = self.hyperparams["train"]["trainer"]["lr"]
        optimizer = optimizers.Adam(learning_rate=lr)

        model.compile(optimizer=optimizer)

        return model

    def create(self, input_name=None):
        model = self.build(input_name)
        model = self.compile(model)

        return model

    @staticmethod
    def transfer_weights(src_model, dest_model):
        weights = src_model.get_weights()
        dest_model.set_weights(weights)

    @staticmethod
    def get_num_params(model):
        return f"{model.count_params():,}"

    @staticmethod
    def load_model(path, compile=False, custom_objects={}):
        return TFmodels.load_model(
            filepath=path,
            compile=compile,
            custom_objects=custom_objects,
        )



"""Job creation for training"""

import os
import stat
import getpass
import itertools
import pandas as pd
from pathlib import Path
from ruamel.yaml import YAML

import tfcaidm.jobs.utils.env as env
import tfcaidm.jobs.utils.config as config
import tfcaidm.common.timedate as timedate
import tfcaidm.jobs.utils.tool as tool


class Jobs(config.Config):
    def __init__(self, path, *args, **kwargs):
        super(Jobs, self).__init__(path=path, *args, **kwargs)

        self.csv_path = None
        self.log_dir = None
        self.script_dir = None
        self.scripts = None
        self.timestamp = None

    def __create_csv(self, params):
        df = pd.DataFrame(params)
        df.to_csv(self.csv_path, index=False)

    def create_folder(self, path):
        Path(path).mkdir(exist_ok=True)

    def __create_folders(self, root="exp", name="unnamed"):
        exp = Path(root).resolve()
        self.create_folder(exp)

        exp = (exp / name).resolve()
        self.create_folder(exp)

        csv_basedir = exp / "csvs"
        log_basedir = exp / "logs"
        script_basedir = exp / "scripts"

        self.create_folder(csv_basedir)
        self.create_folder(log_basedir)
        self.create_folder(script_basedir)

        self.csv_path = str(csv_basedir / self.timestamp) + "_hyper.csv"
        print(f"- csv located at {self.csv_path}")

        self.log_dir = str(log_basedir / self.timestamp)
        self.create_folder(self.log_dir)
        print(f"- logs located at {self.log_dir}")

        self.script_dir = str(script_basedir / self.timestamp)
        self.create_folder(self.script_dir)
        print(f"- scripts located at {self.script_dir}")

    def __create_log_folders(self, params):
        for i, _ in enumerate(params):
            path = str(Path(self.log_dir) / str(i))
            params[i]["env/path/param_csv"] = self.csv_path
            params[i]["train/trainer/log_dir"] = path

            self.create_folder(path)

            param = tool.unflatten_dict(params[i])

            self.create_yml(param, path, ext="pipeline.yml")

    def __create_permutations(self):
        hyperparams = []

        model_params = tool.flatten_dict(self.config["model"])
        train_params = tool.flatten_dict(self.config["train"])

        model_keys = [*model_params.keys()]
        train_keys = [*train_params.keys()]
        model_keys = [f"model/{k}" for k in model_keys]
        train_keys = [f"train/{k}" for k in train_keys]
        keys = model_keys + train_keys

        model_values = [*model_params.values()]
        train_values = [*train_params.values()]
        values = model_values + train_values
        values = [[value] if value is None else value for value in values]

        for i, row in enumerate(itertools.product(*values)):
            params = {}
            params["env/path/root"] = self.config["env"]["path"]["root"]
            params["env/path/name"] = self.config["env"]["path"]["name"]
            params["env/path/client"] = self.config["env"]["path"]["client"]

            params.update(dict(zip(keys, row)))
            hyperparams.append(params)

        return hyperparams

    def __create_scripts(
        self,
        num_gpus,
    ):
        assert num_gpus != 0, "ERROR! Training must be done using at least 1 gpu!"

        if num_gpus < 0:
            num_gpus = len(self.scripts)

        # ------------------------------------------------#
        scripts = {}
        num_models = len(self.scripts)
        leftover = num_models % num_gpus
        allocated = num_models - leftover
        models_per_gpu = allocated // num_gpus

        if allocated == 0:
            allocated = num_models
            models_per_gpu = 1

        counter = 0

        # --- Iterate over all chosen gpus
        for i in range(0, allocated, models_per_gpu):

            # --- Iterate over all models under a given gpu
            contents = ""
            for j in range(models_per_gpu):
                k = i + j
                contents += self.scripts[k]
                contents += "\n"

            scripts[counter] = contents
            counter += 1

        # --- Iterate over remaining models, if any
        if allocated < num_models:
            contents = ""
            for k in range(allocated, num_models):
                contents += self.scripts[k]
                contents += "\n"

            scripts[counter] = contents
        # ------------------------------------------------#

        # --- Populate bash script
        for k, v in scripts.items():

            # --- Write contents to a file
            path = str(Path(self.script_dir) / f"{k}.sh")
            with open(path, "w") as f:
                f.write(v)

            # --- Make script executable
            Jobs.exe(path)

    def create_yml(self, params, path, ext):
        path = str(Path(path) / ext)

        with open(path, "w") as f:
            yaml = YAML()
            yaml.default_flow_style = False
            yaml.dump(params, f)

    def get_params(self):
        """For returning params, useful for local development and debugging"""

        params = self.__create_permutations()
        return params

    def setup(
        self,
        producer,
        consumer,
        root=None,
        name=None,
        libraries=[],
    ):
        assert (Path(producer).name != Path(consumer).name) and (
            Path(producer).resolve().parent == Path(consumer).resolve().parent
        ), f"""ERROR! Please ensure that `producer` and `consumer` are different files within the same directory!

        Issues:
            producer={Path(producer).resolve()}
            consumer={Path(consumer).resolve()}

        Note: producer must be set to __file__ and consumer must point to an external file in the same directory.
        """

        self.timestamp = timedate.get_date()

        params = self.__create_permutations()

        if root is None:
            root = self.config["env"]["path"]["root"]

        if name is None:
            name = self.config["env"]["path"]["name"]

        if (
            len(libraries) == 0
            and "lib" in self.config["env"]
            and self.config["env"]["lib"] is not None
        ):
            libraries = self.config["env"]["lib"]
            libraries = [tuple(v) for k, v in libraries.items()]

        self.__create_folders(root=root, name=name)
        self.__create_log_folders(params)
        self.__create_csv(params)
        self.__setup_scripts(params, consumer=consumer, libraries=libraries)

        return self

    def __setup_scripts(
        self,
        params,
        consumer,
        libraries=[],
    ):
        kwargs = {}
        kwargs["user"] = getpass.getuser()
        kwargs["python_path"] = os.getcwd()
        kwargs["param_path"] = self.csv_path
        kwargs["program"] = str(Path(os.getcwd()) / consumer)
        kwargs["libraries"] = libraries

        scripts = {}

        for i, param in enumerate(params):
            kwargs["log_dir"] = str(Path(self.log_dir) / str(i))
            kwargs["row_id"] = i

            command = env.set_requirements(**kwargs)
            scripts[i] = command

        print(f"- training {i + 1} model configurations")

        self.scripts = scripts

    def train_cluster(self, gpu="titan|rtx", num_gpus=1, run=True):
        """For cluster training (must be invoked in a standalone file)

        Args:
            gpu (str): Name of gpu to run, based on regex matching.
            num_gpus (int): Number of gpus to use for training.
                - Special: a value of -1 means to train all models in parallel.
                           Please be responsible when doing so!
        """

        self.__create_scripts(num_gpus=num_gpus)
        path = str(Path(self.script_dir) / "*.sh")

        if run:
            cmd = (
                "jarvis cluster add -scripts "
                + f'"{path}"'
                + f' -workers "({gpu}).*worker-[0,1]"'
            )
            os.system(cmd)
            print(f"- script {cmd} is running on the clusters.")

    def train_local(self, run=True):
        """For local training training (must be invoked in a standalone file)"""

        self.__create_scripts(num_gpus=1)
        path = str(Path(self.script_dir) / "*.sh")

        if run:
            cmd = f"sh {path}"
            os.system(cmd)
            print(f"- script {cmd} is running locally.")

    @staticmethod
    def exe(path):
        os.chmod(path, stat.S_IEXEC | stat.S_IREAD)



"""Add class weights (for use in loss func)"""

import numpy as np


def vector_weights(xs, ys, hyperparams, condition=True):

    # --- Create a mask template
    msk = np.ones(xs.shape) * (not hyperparams["remove_bg"])

    # --- Binarize masks
    crop = xs > 0
    label = ys > 0

    # --- Set class weights
    msk[crop] = hyperparams["mask_weight"]
    msk[label] = hyperparams[
        "output_weight" if "output_weight" in hyperparams else "mask_weight"
    ]

    return msk * condition


def scalar_weights(xs, ys, hyperparams, condition=True):

    # --- Force mask to 1
    xs[:] = 1 * condition

    return xs


def ignore_row(row, name="cohort-uci"):
    """
    Mask depending on available annotations
    Uci data has ratio no seg, RSNA has seg no ratio
    This is an example for the dual modality covid-biomarker dataset
    """

    if name in row:
        uci_ratio = row[name] == True
        rsna_mask = not uci_ratio
    else:
        uci_ratio = True
        rsna_mask = True

    return uci_ratio, rsna_mask


def modify_loss(xs, ys, row, hyperparams, kwargs):

    n_dim = len(xs.shape)

    uci_ratio, rsna_mask = ignore_row(row)

    # --- Ratio mask
    if n_dim == 1:
        condition = uci_ratio
        return scalar_weights(xs, ys, hyperparams, condition=condition)

    # --- Segmentation mask
    else:
        condition = rsna_mask
        return vector_weights(xs, ys, hyperparams, condition=condition)



"""Defined training callbacks"""

import tfcaidm.train.utils.callbacks as callbacks

calls = {
    "checkpoint": callbacks.model_checkpoints,
    "lr_scheduler": callbacks.lr_scheduler,
    "tensorboard": callbacks.tensorboard_init,
    "exit_on_nan": callbacks.exit_on_nan,
}



"""Tversky related loss functions"""

import tensorflow as tf
from tensorflow.keras import layers


reduce = lambda x: tf.math.reduce_sum(x)


class TverskyLoss(layers.Layer):
    """Computes the tversky (generalized dice) loss"""

    def __init__(self, name="TverskyLoss", **kwargs):
        super(TverskyLoss, self).__init__(name=name, **kwargs)

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        alpha=0.3,
        class_of_interest=1,
        epsilon=1e-6,
        add_loss=True,
        **kwargs,
    ):
        """
        Implementation of Tversky index loss
        TL = tp/(tp + (1-alpha)*fn + alpha*fp) st. alpha = [0, 1]
        = tp/fn when alpha = 0 (punish fn more)
        = tp/(t+p) when alpha = 0.5 (dice score)
        = tp/fp when alpha = 1 (punish fp more)
        """

        # ---- Extract the class of interest
        true = tf.cast(y_true[..., 0] == class_of_interest, tf.float32)
        pred = tf.nn.softmax(y_pred, axis=-1)[..., class_of_interest]

        if weights is not None:
            true *= tf.cast(weights[..., 0], tf.float32)
            pred *= tf.cast(weights[..., 0], tf.float32)

        tp = reduce(true * pred)
        fn = reduce(true * (1 - pred))
        fp = reduce((1 - true) * pred)

        A = tp
        B = tp + ((1 - alpha) * fn) + (alpha * fp) + epsilon

        tt = reduce(true * true)
        ft = reduce(true * (1 - true))

        C = tt
        D = tt + ft + epsilon

        loss = (C / D) - (A / B)

        if add_loss:
            self.add_loss(loss)

        return loss


class FocalTverskyLoss(TverskyLoss):
    """Computes the focal tversky (generalized dice) loss"""

    def __init__(self, name="FocalTverskyLoss", **kwargs):
        super(FocalTverskyLoss, self).__init__(name=name, **kwargs)

    def call(
        self,
        y_true,
        y_pred,
        weights=None,
        alpha=0.3,
        gamma=0.75,
        class_of_interest=1,
        epsilon=1e-6,
        add_loss=True,
        **kwargs,
    ):
        loss = (
            TverskyLoss.call(
                self,
                y_true=y_true,
                y_pred=y_pred,
                weights=weights,
                alpha=alpha,
                class_of_interest=class_of_interest,
                epsilon=epsilon,
                add_loss=False,
            )
            ** gamma
        )

        if add_loss:
            self.add_loss(loss)

        return loss
